"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""


import os
from pathlib import Path

import wx

from LabGym.categorizers import Categorizers, delete_categorizer, get_categorizer_names
from LabGym.gui.utils import BehaviorMode, LabGymWindow

the_absolute_current_path = str(Path(__file__).resolve().parent.parent.parent)


class TrainCategorizers(LabGymWindow):
    """Train a Categorizer model using sorted behavior examples.

    Attributes:
        behavior_example_folder: The location of the sorted behavior examples.
        renamed_example_folder: The location of the renamed, labeled examples.
        behavior_mode: The behavior mode.
        using_animation_analyzer: Whether or not to use animations in training
            the categorizer.
        level_tconv: The depth of the Animation Analyzer.
        level_conv: The depth of the Pattern Recognizer.
        dim_tconv: The width of the Animation Analyzer.
        dim_conv: The width of the Pattern Recognizer.
        channel: The number of channels in ???
        behavior_length: The number of frames corresponding to a single
            behavior.
        aug_methods: A list of augmentation methods to use while training.
        augvalid: Whether or not to use augmentation for validation data.
        data_path: The path to all prepared training examples.
        model_path: The location of the available categorizers.
        path_to_categorizer: The path to the newly created categorizer.
        training_report_path: The folder in which to store training reports.
        include_bodyparts: Whether or not to include body parts.
        std: ???
        example_width: The behavior example width after resizing, if required.
        background_free: Whether or not the animations contain backgrounds.
        social_distance: The social distance when detecting interactive
            behaviors. (Units ???)
    """

    def __init__(self):
        super().__init__(title="Train Categorizers", size=(1200, 550))
        self.behavior_example_folder = None
        self.renamed_example_folder = None
        self.behavior_mode = BehaviorMode.NON_INTERACTIVE
        self.using_animation_analyzer = True
        self.level_tconv = 2
        self.level_conv = 2
        self.dim_tconv = 32
        self.dim_conv = 32
        self.channel = 1
        self.behavior_length = 15
        self.aug_methods = []
        self.augvalid = True
        self.data_path = None
        self.model_path = os.path.join(the_absolute_current_path, "models")
        self.path_to_categorizer = os.path.join(
            the_absolute_current_path, "models", "New_model"
        )
        self.training_report_path = None
        self.include_bodyparts = False
        self.std = 0
        self.example_width = None
        self.background_free = True
        self.social_distance = 0

        self.text_inputexamples = self.module_text("None.")
        self.add_module(
            "Select the folder that stores\nthe sorted behavior examples",
            self.select_behavior_example_folder,
            "This folder should contain all the sorted behavior examples. Each subfolder in this folder should contain behavior examples of a behavior type. The names of the subfolders will be read by LabGym as the behavior names.",
            self.text_inputexamples,
        )

        self.text_renameexample = self.module_text("None.")
        self.add_module(
            "Select a new folder to store\nall the prepared behavior examples",
            self.select_renamed_example_folder,
            "This folder will store all the prepared behavior examples and can be directly used for training. Preparing behavior examples is copying all examples into this folder and renaming them to put behavior name labels to their file names.",
            self.text_renameexample,
        )

        self.add_submit_button(
            "Start to prepare the training examples",
            self.rename_behavior_examples,
            "All prepared behavior examples will be stored in the same folder and ready to be input for training.",
        )

        self.text_categorizertype = self.module_text(
            "Default: Categorizer (Animation Analyzer LV2 + Pattern Recognizer LV2). Behavior mode: Non-interact (identify behavior for each individual)."
        )
        self.add_module(
            "Specify the type / complexity of\nthe Categorizer to train",
            self.configure_categorizer,
            "Categorizer with both Animation Analyzer and Pattern Recognizer is slower but a little more accurate than that with Pattern Recognizer only. Higher complexity level means deeper and more complex network architecture. See Extended Guide for details.",
            self.text_categorizertype,
        )

        self.text_categorizershape = self.module_text(
            "Default: (width,height,channel) is (32,32,1) / (32,32,3)."
        )
        self.add_module(
            "Specify the input shape for\nAnimation Analyzer / Pattern Recognizer",
            self.set_categorizer_shape,
            "The input frame / image size should be an even integer and larger than 8. The larger size, the wider of network architecture. Use large size only when there are detailed features in frames / images that are important for identifying behaviors. See Extended Guide for details.",
            self.text_categorizershape,
        )

        self.text_length = self.module_text("Default: 15.")
        self.add_module(
            "Specify the number of frames for\nan animation / pattern image",
            self.input_behavior_length,
            'The duration (how many frames) of a behavior example. This info can be found in the filenames of the generated behavior examples, "_lenXX_" where "XX" is the number you need to enter here.',
            self.text_length,
        )

        self.text_trainingfolder = self.module_text("None.")
        self.add_module(
            "Select the folder that stores\nall the prepared training examples",
            self.select_training_examples,
            'The folder that stores all the prepared behavior examples. If body parts are included, the STD value can be found in the filenames of the generated behavior examples with "_stdXX_" where "XX" is the number you need to enter here.',
            self.text_trainingfolder,
        )

        self.text_augmentation = self.module_text("None.")
        self.add_module(
            "Specify the methods to\naugment training examples",
            self.specify_augmentation,
            'Randomly change or add noise into the training examples to increase their amount and diversity, which can benefit the training. If the amount of examples less than 1,000 before augmentation, choose "Also augment the validation data". See Extended Guide for details.',
            self.text_augmentation,
        )

        self.text_report = self.module_text("None.")
        self.add_module(
            "Select a folder to\nexport training reports",
            self.select_report_path,
            "This is the folder to store the reports of training history and metrics. It is optional.",
            self.text_report,
        )

        self.add_submit_button(
            "Train the Categorizer",
            self.train_categorizer,
            "Need to name the Categorizer to train. English letters, numbers, underscore “_”, or hyphen “-” are acceptable but do not use special characters such as “@” or “^”.",
        )

        self.display_window()

    def select_behavior_example_folder(self, event):
        """Select the folder containing sorted behavior examples."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.behavior_example_folder = dialog.GetPath()
            self.text_inputexamples.SetLabel(
                f"Path to sorted behavior examples: {self.behavior_example_folder}."
            )
        dialog.Destroy()

    def select_renamed_example_folder(self, event):
        """Select directory to store renamed behavior examples."""
        dialog = wx.DirDialog(
            self, "Select a new directory", "", style=wx.DD_DEFAULT_STYLE
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.renamed_example_folder = dialog.GetPath()
            self.text_renameexample.SetLabel(
                f"Will copy and rename the examples to: {self.renamed_example_folder}."
            )
        dialog.Destroy()

        dialog = wx.MessageDialog(
            self,
            'Reducing frame / image size can speed up training\nSelect "No" if dont know what it is.',
            "Resize the frames / images?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if dialog.ShowModal() != wx.ID_YES:
            self.example_width = None
            dialog.Destroy()
            return

        width_dialog = wx.NumberEntryDialog(
            self,
            "Enter the desired width dimension",
            "No smaller than the\ndesired input dimension of the Categorizer:",
            "Frame / image dimension",
            32,
            1,
            300,
        )
        if width_dialog.ShowModal() != wx.ID_OK:
            width_dialog.Destroy()
            return
        else:
            self.example_width = max(int(width_dialog.GetValue()), 16)
            width_dialog.Destroy()

        self.text_renameexample.SetLabel(
            f"Will copy, rename, and resize (to {self.example_width}) the examples to {self.renamed_example_folder}."
        )

    def rename_behavior_examples(self, event):
        """Move and rename behavior examples for use with Categorizers."""
        if self.behavior_example_folder is None or self.renamed_example_folder is None:
            wx.MessageBox(
                "Please select a folder that stores the sorted examples /\na new folder to store prepared training examples!",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
        else:
            CA = Categorizers()
            CA.rename_label(
                self.behavior_example_folder,
                self.renamed_example_folder,
                resize=self.example_width,
            )

    def configure_categorizer(self, event):
        """Configure the Categorizer."""

        # These are mapped to the same constants in the BehaviorMode class.
        behavior_modes = [
            "Non-interact (identify behavior for each individual)",
            "Interact basic (identify behavior for the interactive pair / group)",
            "Interact advanced (identify behavior for both each individual and each interactive pair / group)",
            "Static images (non-interactive): behaviors of each individual in static images(each image contains one animal / object)",
        ]
        dialog = wx.SingleChoiceDialog(
            self,
            message="Specify the mode of behavior for the Categorizer to identify",
            caption="Behavior mode",
            choices=behavior_modes,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        else:
            self.behavior_mode = dialog.GetSelection()
            dialog.Destroy()

        if self.behavior_mode == BehaviorMode.INTERACT_ADVANCED:
            self.channel = 3
            distance_dialog = wx.NumberEntryDialog(
                self,
                "Interactions happen within the social distance.",
                "How many folds of the animals's diameter\nis the social distance (0=inf):",
                "Social distance (Enter an integer)",
                0,
                0,
                100000000000000,
            )
            if distance_dialog.ShowModal() != wx.ID_OK:
                distance_dialog.Destroy()
                self.social_distance = float("inf")
                return
            else:
                self.social_distance = float(distance_dialog.GetValue())
                if self.social_distance == 0:
                    self.social_distance = float("inf")
                distance_dialog.Destroy()
        elif self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            self.text_length.SetLabel(
                'No need to specify this since the selected behavior mode is "Static images".'
            )

        PATTERN_RECOGNIZER_ONLY = (
            "Categorizer (Pattern Recognizer only) (faster / a little less accurate)"
        )
        BOTH = "Categorizer (Animation Analyzer + Pattern Recognizer)"
        categorizer_types = [PATTERN_RECOGNIZER_ONLY]
        if self.behavior_mode != BehaviorMode.STATIC_IMAGES:
            categorizer_types.append(BOTH)

        dialog = wx.SingleChoiceDialog(
            self,
            message="Select the Categorizer type",
            caption="Categorizer types",
            choices=categorizer_types,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        else:
            categorizer_tp = dialog.GetStringSelection()
            self.using_animation_analyzer = categorizer_tp != PATTERN_RECOGNIZER_ONLY
            dialog.Destroy()

        dialog1 = wx.NumberEntryDialog(
            self,
            "Complexity level from 1 to 7\nhigher level = deeper network",
            "Enter a number (1~7)",
            "Pattern Recognizer level",
            2,
            1,
            7,
        )
        if dialog1.ShowModal() != wx.ID_OK:
            dialog1.Destroy()
            return
        else:
            self.level_conv = int(dialog1.GetValue())
            dialog1.Destroy()

        if self.using_animation_analyzer:
            dialog1 = wx.NumberEntryDialog(
                self,
                "Complexity level from 1 to 7\nhigher level = deeper network",
                "Enter a number (1~7)",
                "Animation Analyzer level",
                2,
                1,
                7,
            )
            if dialog1.ShowModal() != wx.ID_OK:
                dialog1.Destroy()
                return
            else:
                self.level_tconv = int(dialog1.GetValue())
                dialog1.Destroy()

        label_text = f"Using Pattern Recognizer (Level {self.level_conv})"
        if self.using_animation_analyzer:
            label_text += f" and Animation Analyzer (Level {self.level_tconv})"

        if self.behavior_mode == BehaviorMode.NON_INTERACTIVE:
            label_text += " to identify non-interactive behaviors of each individual."
        elif self.behavior_mode == BehaviorMode.INTERACT_BASIC:
            label_text += " to identify interactive behaviors of the group."
        elif self.behavior_mode == BehaviorMode.INTERACT_ADVANCED:
            label_text += (
                " to identify interactive behaviors of the individual and groups."
            )
            label_text += (
                f" Social Distance: {self.social_distance} folds of animal diameter."
            )
        else:
            label_text += " to identify non-interactive behaviors of each individual in static images."

        self.text_categorizertype.SetLabel(label_text)

    def set_categorizer_shape(self, event):
        """Set the network shape of the categorizer."""
        if self.using_animation_analyzer is True:
            dialog = wx.NumberEntryDialog(
                self,
                "Input dimension of Animation Analyzer\nlarger dimension = wider network",
                "Enter a number:",
                "Animation Analyzer input",
                32,
                1,
                300,
            )
            if dialog.ShowModal() != wx.ID_OK:
                dialog.Destroy()
                return
            else:
                self.dim_tconv = int(dialog.GetValue())
                dialog.Destroy()

            self.channel = 3
            if self.behavior_mode != BehaviorMode.INTERACT_ADVANCED:
                dialog = wx.MessageDialog(
                    self,
                    'Grayscale input of Animation Analyzer?\nSelect "Yes" if the color of animals is behavior irrelevant.',
                    "Grayscale Animation Analyzer?",
                    wx.YES_NO | wx.ICON_QUESTION,
                )
                if dialog.ShowModal() == wx.ID_YES:
                    self.channel = 1
                dialog.Destroy()

        dialog = wx.NumberEntryDialog(
            self,
            "Input dimension of Pattern Recognizer\nlarger dimension = wider network",
            "Enter a number:",
            "Input the dimension",
            32,
            1,
            300,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        else:
            self.dim_conv = int(dialog.GetValue())
            dialog.Destroy()

        self.channel = 3
        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            dialog1 = wx.MessageDialog(
                self,
                'Grayscale input?\nSelect "Yes" if the color of animals is behavior irrelevant.',
                "Grayscale inputs?",
                wx.YES_NO | wx.ICON_QUESTION,
            )
            if dialog1.ShowModal() == wx.ID_YES:
                self.channel = 1
            dialog1.Destroy()

        shape_tconv = f"({self.dim_tconv}, {self.dim_tconv}, {self.channel})"
        shape_conv = f"({self.dim_conv}, {self.dim_conv}, {self.channel})"

        label_text = f"Input shapes: Pattern Recognizer {shape_conv}"
        if self.using_animation_analyzer:
            label_text += f", Animation Analyzer {shape_tconv}."
        else:
            label_text += "."
        self.text_categorizershape.SetLabel(label_text)

    def input_behavior_length(self, event):
        """Set the number of frames corresponding to a behavior."""
        if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        dialog = wx.NumberEntryDialog(
            self,
            "The number of frames of\na behavior example",
            "Enter a number (minimum=3):",
            "Behavior episode duration",
            15,
            1,
            1000,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.behavior_length = max(int(dialog.GetValue()), 3)
            self.text_length.SetLabel(
                f"The duration of a behavior example is: {self.behavior_length}."
            )
        dialog.Destroy()

    def select_training_examples(self, event):
        """Select the directory with the training examples."""
        dialog = wx.MessageDialog(
            self,
            "Are the animations (in any) in\ntraining examples background free?",
            "Background-free animations?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        self.background_free = dialog.ShowModal == wx.ID_YES
        dialog.Destroy()

        self.include_bodyparts = False
        if self.behavior_mode != BehaviorMode.STATIC_IMAGES:
            dialog = wx.MessageDialog(
                self,
                "Do the pattern images in training examples\ninclude body parts?",
                "Body parts in pattern images?",
                wx.YES_NO | wx.ICON_QUESTION,
            )
            if dialog.ShowModal() == wx.ID_YES:
                self.include_bodyparts = True
                dialog2 = wx.NumberEntryDialog(
                    self,
                    "Should match the STD of the pattern images in training examples.",
                    "Enter a number between 0 and 255:",
                    "STD for motionless pixels",
                    0,
                    0,
                    255,
                )
                if dialog2.ShowModal() == wx.ID_OK:
                    self.std = int(dialog2.GetValue())
                else:
                    self.std = 0
                dialog2.Destroy()
            dialog.Destroy()

        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        else:
            self.data_path = dialog.GetPath()
            dialog.Destroy()

        label_text = "Animations w/"
        if self.background_free:
            label_text += "o"
        label_text += " background, "
        if self.include_bodyparts:
            label_text += f"pattern images w/ bodyparts (std={self.std})"
        else:
            label_text += "pattern images w/o bodyparts"
        label_text += f" in: {self.data_path}."
        self.text_trainingfolder.SetLabel(label_text)

    def specify_augmentation(self, event):
        """Select augmentation methods."""
        dialog = wx.MessageDialog(
            self,
            'Use default augmentation methods?\nSelect "Yes" if dont know how to specify.',
            "Use default augmentation?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if dialog.ShowModal() == wx.ID_YES:
            self.aug_methods = [
                "random rotation",
                "horizontal flipping",
                "vertical flipping",
                "random brightening",
                "random dimming",
            ]
        else:
            aug_methods = [
                "random rotation",
                "horizontal flipping",
                "vertical flipping",
                "random brightening",
                "random dimming",
                "random shearing",
                "random rescaling",
                "random deletion",
            ]
            dialog1 = wx.MultiChoiceDialog(
                self,
                message="Data augmentation methods",
                caption="Augmentation methods",
                choices=aug_methods,
            )
            self.aug_methods = []
            if dialog1.ShowModal() == wx.ID_OK:
                self.aug_methods = [aug_methods[i] for i in dialog1.GetSelections()]
            dialog1.Destroy()
        dialog.Destroy()

        dialog = wx.MessageDialog(
            self,
            'Also augment the validation data?\nSelect "No" if dont know what it is.',
            "Augment validation data?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if dialog.ShowModal() == wx.ID_YES:
            self.augvalid = True
            self.text_augmentation.SetLabel(
                f"Augment both training and validation examples with: {', '.join(self.aug_methods) if self.aug_methods else 'none'}"
            )
        else:
            self.augvalid = False
            self.text_augmentation.SetLabel(
                f"Augment training examples with: {', '.join(self.aug_methods) if self.aug_methods else 'none'}"
            )
        dialog.Destroy()

    def select_report_path(self, event):
        dialog = wx.MessageDialog(
            self,
            "Export the training reports?",
            "Export training reports?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if dialog.ShowModal() == wx.ID_YES:
            dialog2 = wx.DirDialog(
                self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE
            )
            if dialog2.ShowModal() == wx.ID_OK:
                self.training_report_path = dialog2.GetPath()
                self.text_report.SetLabel(
                    f"Training reports will be in: {self.training_report_path}."
                )
            dialog2.Destroy()
        else:
            self.training_report_path = None
        dialog.Destroy()

    def train_categorizer(self, event):
        if self.data_path is None:
            wx.MessageBox(
                "No path to training examples.", "Error", wx.OK | wx.ICON_ERROR
            )
            return

        while True:
            dialog = wx.TextEntryDialog(
                self,
                "Enter a name for the Categorizer to train",
                "Categorizer name",
            )
            if dialog.ShowModal() != wx.ID_OK:
                break

            if dialog.GetValue().strip() == "":
                wx.MessageBox(
                    "Please enter a name.",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                continue
            categorizer_name = dialog.GetValue()
            dialog.Destroy()
            if categorizer_name in get_categorizer_names():
                wx.MessageBox(
                    f"The Categorizer {categorizer_name} already exists!",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                continue

            self.path_to_categorizer = os.path.join(self.model_path, dialog.GetValue())
            os.makedirs(self.path_to_categorizer)

            if self.using_animation_analyzer is False:
                CA = Categorizers()
                if self.behavior_mode == BehaviorMode.STATIC_IMAGES:
                    self.behavior_length = self.std = 0
                    self.include_bodyparts = False
                else:
                    self.channel = 3
                CA.train_pattern_recognizer(
                    self.data_path,
                    self.path_to_categorizer,
                    self.training_report_path,
                    dim=self.dim_conv,
                    channel=self.channel,
                    time_step=self.behavior_length,
                    level=self.level_conv,
                    aug_methods=self.aug_methods,
                    augvalid=self.augvalid,
                    include_bodyparts=self.include_bodyparts,
                    std=self.std,
                    background_free=self.background_free,
                    behavior_mode=self.behavior_mode,
                    social_distance=self.social_distance,
                )
            else:
                if self.behavior_mode == BehaviorMode.INTERACT_ADVANCED:
                    self.channel = 3
                CA = Categorizers()
                CA.train_combnet(
                    self.data_path,
                    self.path_to_categorizer,
                    self.training_report_path,
                    dim_tconv=self.dim_tconv,
                    dim_conv=self.dim_conv,
                    channel=self.channel,
                    time_step=self.behavior_length,
                    level_tconv=self.level_tconv,
                    level_conv=self.level_conv,
                    aug_methods=self.aug_methods,
                    augvalid=self.augvalid,
                    include_bodyparts=self.include_bodyparts,
                    std=self.std,
                    background_free=self.background_free,
                    behavior_mode=self.behavior_mode,
                    social_distance=self.social_distance,
                )
            break


class TestCategorizers(LabGymWindow):
    """Test a categorizer."""

    def __init__(self):
        super().__init__(title="Test Categorizers", size=(1000, 240))
        self.file_path = None
        self.model_path = os.path.join(the_absolute_current_path, "models")
        self.path_to_categorizer = None
        self.out_path = None

        self.text_selectcategorizer = self.module_text("None.")
        self.add_module(
            "Select a Categorizer\nto test",
            self.select_categorizer,
            "The behavioral names in ground-truth dataset should exactly match those in the selected Categorizer.",
            self.text_selectcategorizer,
        )

        self.text_inputexamples = self.module_text("None.")
        self.add_module(
            "Select the folder that stores\nthe ground-truth behavior examples",
            self.select_filepath,
            "This folder should contain all the sorted behavior examples. Each subfolder in this folder should contain behavior examples of a behavior type. The names of the subfolders are the ground-truth behavior names, which should match those in the selected Categorizer.",
            self.text_inputexamples,
        )

        self.text_report = self.module_text("None.")
        self.add_module(
            "Select a folder to\nexport testing reports",
            self.select_reportpath,
            "This is the folder to store the reports of testing results and metrics. It is optional.",
            self.text_report,
        )

        test_and_delete = wx.BoxSizer(wx.HORIZONTAL)
        button_test = wx.Button(
            self.panel, label="Test the Categorizer", size=(300, 40)
        )
        button_test.Bind(wx.EVT_BUTTON, self.test_categorizer)
        wx.Button.SetToolTip(
            button_test,
            "Test the selected Categorizer on the ground-truth behavior examples",
        )

        button_delete = wx.Button(
            self.panel, label="Delete a Categorizer", size=(300, 40)
        )
        button_delete.Bind(wx.EVT_BUTTON, self.remove_categorizer)
        wx.Button.SetToolTip(
            button_delete,
            "Permanently delete a Categorizer. The deletion CANNOT be restored.",
        )
        test_and_delete.Add(button_test, 0, wx.RIGHT, 50)
        test_and_delete.Add(button_delete, 0, wx.LEFT, 50)
        self.boxsizer.Add(0, 5, 0)
        self.boxsizer.Add(test_and_delete, 0, wx.RIGHT | wx.ALIGN_RIGHT, 90)
        self.boxsizer.Add(0, 10, 0)

        self.display_window()

    def select_categorizer(self, event):
        """Select the categorizer to test."""
        categorizers = get_categorizer_names()
        LOAD_NEW_CATEGORIZER = "Choose a new directory of the Categorizer"
        categorizers.append(LOAD_NEW_CATEGORIZER)

        dialog = wx.SingleChoiceDialog(
            self,
            message="Select a Categorizer to test",
            caption="Select a Categorizer",
            choices=categorizers,
        )

        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        else:
            categorizer = dialog.GetStringSelection()
            dialog.Destroy()

        if categorizer == LOAD_NEW_CATEGORIZER:
            dialog1 = wx.DirDialog(
                self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE
            )
            if dialog1.ShowModal() != wx.ID_OK:
                dialog1.Destroy()
                return
            else:
                self.path_to_categorizer = dialog1.GetPaths()
                dialog1.Destroy()

            self.text_selectcategorizer.SetLabel(
                f"The path to the Categorizer to test is: {self.path_to_categorizer}."
            )
        else:
            self.path_to_categorizer = os.path.join(self.model_path, categorizer)
            self.text_selectcategorizer.SetLabel(f"Categorizer to test: {categorizer}.")

    def select_filepath(self, event):
        """Select the directory with ground-truth behavior examples."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.file_path = dialog.GetPath()
            self.text_inputexamples.SetLabel(
                f"Path to ground-truth behavior examples: {self.file_path}."
            )
        dialog.Destroy()

    def select_reportpath(self, event):
        """Select the path to store testing reports."""
        dialog = wx.MessageDialog(
            self,
            "Export the testing reports?",
            "Export testing reports?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if dialog.ShowModal() == wx.ID_YES:
            dialog1 = wx.DirDialog(
                self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE
            )
            if dialog1.ShowModal() == wx.ID_OK:
                self.out_path = dialog1.GetPath()
                self.text_report.SetLabel(
                    f"Testing reports will be in: {self.out_path}."
                )
            dialog1.Destroy()
        else:
            self.out_path = None
        dialog.Destroy()

    def test_categorizer(self, event):
        """Test a categorizer."""
        if self.file_path is None or self.path_to_categorizer is None:
            wx.MessageBox(
                "No Categorizer selected / path to ground-truth behavior examples.",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
        else:
            CA = Categorizers()
            CA.test_categorizer(
                self.file_path, self.path_to_categorizer, result_path=self.out_path
            )

    def remove_categorizer(self, event):
        """Delete the a categorizer."""
        categorizers = get_categorizer_names()

        dialog = wx.SingleChoiceDialog(
            self,
            message="Select a Categorizer to delete",
            caption="Delete a Categorizer",
            choices=categorizers,
        )
        if dialog.ShowModal() == wx.ID_OK:
            categorizer = dialog.GetStringSelection()
            dialog1 = wx.MessageDialog(
                self,
                f"Delete {categorizer}?",
                "CANNOT be restored!",
                wx.YES_NO | wx.ICON_QUESTION,
            )
            if dialog1.ShowModal() == wx.ID_YES:
                delete_categorizer(categorizer)
            dialog1.Destroy()
        dialog.Destroy()



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext.

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""

import json
import os
from pathlib import Path

import matplotlib as mpl
import pandas as pd
import torch
import wx

from LabGym.analyzebehaviors import AnalyzeAnimal
from LabGym.analyzebehaviorsdetector import AnalyzeAnimalDetector
from LabGym.gui.utils import WX_IMAGE_WILDCARD, WX_VIDEO_WILDCARD, LabGymWindow
from LabGym.tools import plot_evnets

the_absolute_current_path = str(Path(__file__).resolve().parent.parent.parent)


class ColorPicker(wx.Dialog):
    def __init__(self, parent, title, name_and_color):
        super(ColorPicker, self).__init__(parent=None, title=title, size=(200, 200))

        self.name_and_color = name_and_color
        name = self.name_and_color[0]
        hex_color = self.name_and_color[1][1].lstrip("#")
        color = tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))

        boxsizer = wx.BoxSizer(wx.VERTICAL)

        self.color_picker = wx.ColourPickerCtrl(self, colour=color)

        button = wx.Button(self, wx.ID_OK, label="Apply")

        boxsizer.Add(0, 10, 0)
        boxsizer.Add(self.color_picker, 0, wx.ALL | wx.CENTER, 10)
        boxsizer.Add(button, 0, wx.ALL | wx.CENTER, 10)
        boxsizer.Add(0, 10, 0)

        self.SetSizer(boxsizer)


class AnalyzeBehaviors(LabGymWindow):
    def __init__(self):
        super().__init__(title="Analyze Behaviors", size=(1000, 600))
        self.behavior_mode = 0
        self.use_detector = False
        self.detector_path = None
        self.path_to_detector = None
        self.detector_batch = 1
        self.detection_threshold = 0
        self.animal_kinds = []
        self.background_path = (
            None  # if not None, will load background images from path
        )
        self.model_path = None  # the parent path of the Categorizers
        self.path_to_categorizer = None
        self.path_to_videos = None
        self.result_path = None
        self.framewidth = None
        self.delta = 10000
        self.decode_animalnumber = False
        self.animal_number = None
        self.autofind_t = False
        self.decode_t = False
        self.t = 0
        self.duration = 0
        self.decode_extraction = False
        self.ex_start = 0
        self.ex_end = None
        self.behaviornames_and_colors = {}
        self.dim_tconv = 8
        self.dim_conv = 8
        self.channel = 1
        self.length = 15
        self.animal_vs_bg = 0  # 0: animals birghter than the background; 1: animals darker than the background; 2: hard to tell
        self.stable_illumination = True
        self.animation_analyzer = True
        self.animal_to_include = []
        self.behavior_to_include = ["all"]  # behaviors for annotation and analyze
        self.parameter_to_analyze = []
        self.include_bodyparts = False
        self.std = 0
        self.uncertain = 0
        self.show_legend = True
        self.background_free = True
        self.normalize_distance = True  # whether to normalize the distance (in pixel) to the animal contour area
        self.social_distance = 0
        self.specific_behaviors = {}
        self.correct_ID = False

        self.dispaly_window()

    def dispaly_window(self):
        self.text_selectcategorizer = self.module_text(
            "Default: no behavior classification, just track animals and quantify motion kinematcis."
        )
        self.add_module(
            button_label="Select a Categorizer for\nbehavior classification",
            button_handler=self.select_categorizer,
            tool_tip="The fps of the videos to analyze should match that of the selected Categorizer. Uncertain level determines the threshold for the Categorizer to output an ‘NA’ for behavioral classification. See Extended Guide for details.",
            text=self.text_selectcategorizer,
        )

        self.text_inputvideos = self.module_text("None.")
        self.add_module(
            button_label="Select the video(s) / image(s)\nfor behavior analysis",
            button_handler=self.select_videos,
            tool_tip='Select one or more videos / images for a behavior analysis batch. If analyzing videos, one analysis batch will yield one raster plot showing the behavior events of all the animals in all selected videos. For "Static images" mode, each annotated images will be in this folder. See Extended Guide for details.',
            text=self.text_inputvideos,
        )

        self.text_outputfolder = self.module_text("None.")
        self.add_module(
            button_label="Select a folder to store\nthe analysis results",
            button_handler=self.select_outpath,
            tool_tip='If analyzing videos, will create a subfolder for each video in the selected folder. Each subfolder is named after the file name of the video and stores the detailed analysis results for this video. For "Static images" mode, all results will be in this folder. See Extended Guide for details.',
            text=self.text_outputfolder,
        )

        self.text_detection = self.module_text(
            "Default: Background subtraction-based method."
        )
        self.add_module(
            button_label="Specify the method to\ndetect animals or objects",
            button_handler=self.select_method,
            tool_tip="Background subtraction-based method is accurate and fast but requires static background and stable illumination in videos; Detectors-based method is accurate and versatile in any recording settings but is slow. See Extended Guide for details.",
            text=self.text_detection,
        )

        self.text_startanalyze = self.module_text(
            "Default: at the beginning of the video(s)."
        )
        self.add_module(
            button_label="Specify when the analysis\nshould begin (unit: second)",
            button_handler=self.specify_timing,
            tool_tip='Enter a beginning time point for all videos in one analysis batch or use "Decode from filenames" to let LabGym decode the different beginning time for different videos. See Extended Guide for details.',
            text=self.text_startanalyze,
        )

        self.text_duration = self.module_text(
            "Default: from the specified start time to the end of a video"
        )
        self.add_module(
            button_label="Specify the analysis duration\n(unit: second)",
            button_handler=self.input_duration,
            tool_tip="The duration is the same for all the videos in a same analysis batch.",
            text=self.text_duration,
        )

        self.text_animalnumber = self.module_text("Default: 1.")
        self.add_module(
            button_label="Specify the number of animals\nin a video",
            button_handler=self.specify_animalnumber,
            tool_tip='Enter a number for all videos in one analysis batch or use "Decode from filenames" to let LabGym decode the different animal number for different videos. See Extended Guide for details.',
            text=self.text_animalnumber,
        )

        self.text_selectbehaviors = self.module_text(
            "Default: No Categorizer selected, no behavior selected."
        )
        self.add_module(
            button_label="Select the behaviors for\nannotations and plots",
            button_handler=self.select_behaviors,
            tool_tip="The behavior categories are determined by the selected Categorizer. Select which behaviors to show in the annotated videos / images and the raster plot (only for videos). See Extended Guide for details.",
            text=self.text_selectbehaviors,
        )

        self.text_selectparameters = self.module_text("Default: None.")
        self.add_module(
            button_label="Select the quantitative measurements\nfor each behavior",
            button_handler=self.select_parameters,
            tool_tip='If select "not to normalize distances", all distances will be output in pixels. If select "normalize distances", all distances will be normalized to the animal size. See Extended Guide for details.',
            text=self.text_selectparameters,
        )

        self.add_submit_button(
            label="Start to analyze the behaviors",
            handler=self.analyze_behaviors,
            tool_tip='If analyzing videos, will output a raster plot for all behavior events in all videos, an annotated video copy for each video, various spreadsheets storing quantification results for each selected behavior parameter. For "Static images" mode, will output annotated image copies and spreadsheet storing behavior count and probability.',
        )

        self.display_window()

    def select_categorizer(self, event):
        if self.model_path is None:
            self.model_path = os.path.join(the_absolute_current_path, "models")

        categorizers = [
            i
            for i in os.listdir(self.model_path)
            if os.path.isdir(os.path.join(self.model_path, i))
        ]
        if "__pycache__" in categorizers:
            categorizers.remove("__pycache__")
        if "__init__" in categorizers:
            categorizers.remove("__init__")
        if "__init__.py" in categorizers:
            categorizers.remove("__init__.py")
        categorizers.sort()
        if (
            "No behavior classification, just track animals and quantify motion kinematics"
            not in categorizers
        ):
            categorizers.append(
                "No behavior classification, just track animals and quantify motion kinematics"
            )
        if "Choose a new directory of the Categorizer" not in categorizers:
            categorizers.append("Choose a new directory of the Categorizer")

        dialog = wx.SingleChoiceDialog(
            self,
            message="Select a Categorizer for behavior classification",
            caption="Select a Categorizer",
            choices=categorizers,
        )

        if dialog.ShowModal() == wx.ID_OK:
            categorizer = dialog.GetStringSelection()
            if categorizer == "Choose a new directory of the Categorizer":
                dialog1 = wx.DirDialog(
                    self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE
                )
                if dialog1.ShowModal() == wx.ID_OK:
                    self.path_to_categorizer = dialog1.GetPaths()
                dialog1.Destroy()
                dialog1 = wx.NumberEntryDialog(
                    self,
                    "Enter the Categorizer's uncertainty level (0~100%)",
                    "If probability difference between\n1st- and 2nd-likely behaviors\nis less than uncertainty,\nclassfication outputs an 'NA'. Enter 0 if don't know how to set.",
                    "Uncertainty level",
                    0,
                    0,
                    100,
                )
                if dialog1.ShowModal() == wx.ID_OK:
                    uncertain = dialog1.GetValue()
                    self.uncertain = uncertain / 100
                dialog1.Destroy()
                self.text_selectcategorizer.SetLabel(
                    "The path to the Categorizer is: "
                    + self.path_to_categorizer
                    + " with uncertainty of "
                    + str(uncertain)
                    + "%."
                )
            elif (
                categorizer
                == "No behavior classification, just track animals and quantify motion kinematics"
            ):
                self.path_to_categorizer = None
                dialog1 = wx.NumberEntryDialog(
                    self,
                    "Specify a time window used for measuring\nmotion kinematics of the tracked animals",
                    "Enter the number of\nframes (minimum=3):",
                    "Time window for calculating kinematics",
                    15,
                    1,
                    100000000000000,
                )
                if dialog1.ShowModal() == wx.ID_OK:
                    self.length = int(dialog1.GetValue())
                    if self.length < 3:
                        self.length = 3
                dialog1.Destroy()
                self.text_selectcategorizer.SetLabel(
                    "No behavior classification; the time window to measure kinematics of tracked animals is: "
                    + str(self.length)
                    + " frames."
                )
                self.text_selectbehaviors.SetLabel(
                    "No behavior classification. Just track animals and quantify motion kinematics."
                )
            else:
                self.path_to_categorizer = os.path.join(self.model_path, categorizer)
                dialog1 = wx.NumberEntryDialog(
                    self,
                    "Enter the Categorizer's uncertainty level (0~100%)",
                    "If probability difference between\n1st- and 2nd-likely behaviors\nis less than uncertainty,\nclassfication outputs an 'NA'.",
                    "Uncertainty level",
                    0,
                    0,
                    100,
                )
                if dialog1.ShowModal() == wx.ID_OK:
                    uncertain = dialog1.GetValue()
                    self.uncertain = uncertain / 100
                    self.text_selectcategorizer.SetLabel(
                        "Categorizer: "
                        + categorizer
                        + " with uncertainty of "
                        + str(uncertain)
                        + "%."
                    )
            self.text_selectbehaviors.SetLabel(
                "All the behaviors in the selected Categorizer with default colors."
            )

            if self.path_to_categorizer is not None:
                parameters = pd.read_csv(
                    os.path.join(self.path_to_categorizer, "model_parameters.txt")
                )
                complete_colors = list(mpl.colors.cnames.values())
                colors = []
                for c in complete_colors:
                    colors.append(["#ffffff", c])
                self.behaviornames_and_colors = {}

                for behavior_name in list(parameters["classnames"]):
                    index = list(parameters["classnames"]).index(behavior_name)
                    if index < len(colors):
                        self.behaviornames_and_colors[behavior_name] = colors[index]
                    else:
                        self.behaviornames_and_colors[behavior_name] = [
                            "#ffffff",
                            "#ffffff",
                        ]

                if "dim_conv" in parameters:
                    self.dim_conv = int(parameters["dim_conv"][0])
                if "dim_tconv" in parameters:
                    self.dim_tconv = int(parameters["dim_tconv"][0])
                self.channel = int(parameters["channel"][0])
                self.length = int(parameters["time_step"][0])
                if self.length < 3:
                    self.length = 3
                categorizer_type = int(parameters["network"][0])
                if categorizer_type == 2:
                    self.animation_analyzer = True
                else:
                    self.animation_analyzer = False
                if int(parameters["inner_code"][0]) == 0:
                    self.include_bodyparts = True
                else:
                    self.include_bodyparts = False
                self.std = int(parameters["std"][0])
                if int(parameters["background_free"][0]) == 0:
                    self.background_free = True
                else:
                    self.background_free = False
                if "behavior_kind" in parameters:
                    self.behavior_mode = int(parameters["behavior_kind"][0])
                else:
                    self.behavior_mode = 0
                if self.behavior_mode == 2:
                    self.social_distance = int(parameters["social_distance"][0])
                    if self.social_distance == 0:
                        self.social_distance = float("inf")
                    self.text_detection.SetLabel(
                        "Only Detector-based detection method is available for the selected Categorizer."
                    )
                if self.behavior_mode == 3:
                    self.text_detection.SetLabel(
                        "Only Detector-based detection method is available for the selected Categorizer."
                    )
                    self.text_startanalyze.SetLabel(
                        'No need to specify this since the selected behavior mode is "Static images".'
                    )
                    self.text_duration.SetLabel(
                        'No need to specify this since the selected behavior mode is "Static images".'
                    )
                    self.text_animalnumber.SetLabel(
                        'No need to specify this since the selected behavior mode is "Static images".'
                    )
                    self.text_selectparameters.SetLabel(
                        'No need to specify this since the selected behavior mode is "Static images".'
                    )

        dialog.Destroy()

    def select_videos(self, event):
        if self.behavior_mode == 3:
            wildcard = WX_IMAGE_WILDCARD
        else:
            wildcard = WX_VIDEO_WILDCARD

        dialog = wx.FileDialog(
            self, "Select video(s) / image(s)", "", "", wildcard, style=wx.FD_MULTIPLE
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.path_to_videos = dialog.GetPaths()
            self.path_to_videos.sort()
            path = os.path.dirname(self.path_to_videos[0])
            dialog1 = wx.MessageDialog(
                self,
                'Proportional resize the video frames / images? Reducing frame / image size\nis highly recommended. But select "No" if dont know what it is.',
                "(Optional) resize the frames / images?",
                wx.YES_NO | wx.ICON_QUESTION,
            )
            if dialog1.ShowModal() == wx.ID_YES:
                dialog2 = wx.NumberEntryDialog(
                    self,
                    "Enter the desired frame / image width",
                    "The unit is pixel:",
                    "Desired frame / image width",
                    480,
                    1,
                    10000,
                )
                if dialog2.ShowModal() == wx.ID_OK:
                    self.framewidth = int(dialog2.GetValue())
                    if self.framewidth < 10:
                        self.framewidth = 10
                    self.text_inputvideos.SetLabel(
                        "Selected "
                        + str(len(self.path_to_videos))
                        + " file(s) in: "
                        + path
                        + " (proportionally resize frame / image width to "
                        + str(self.framewidth)
                        + ")."
                    )
                else:
                    self.framewidth = None
                    self.text_inputvideos.SetLabel(
                        "Selected "
                        + str(len(self.path_to_videos))
                        + " file(s) in: "
                        + path
                        + " (original frame / image size)."
                    )
                dialog2.Destroy()
            else:
                self.framewidth = None
                self.text_inputvideos.SetLabel(
                    "Selected "
                    + str(len(self.path_to_videos))
                    + " file(s) in: "
                    + path
                    + " (original frame / image size)."
                )
            dialog1.Destroy()
        dialog.Destroy()

    def select_outpath(self, event):
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.result_path = dialog.GetPath()
            self.text_outputfolder.SetLabel(
                "Results will be in: " + self.result_path + "."
            )
        dialog.Destroy()

    def select_method(self, event):
        if self.behavior_mode <= 1:
            methods = [
                "Subtract background (fast but requires static background & stable illumination)",
                "Use trained Detectors (versatile but slow)",
            ]
        else:
            methods = ["Use trained Detectors (versatile but slow)"]

        dialog = wx.SingleChoiceDialog(
            self,
            message="How to detect the animals?",
            caption="Detection methods",
            choices=methods,
        )

        if dialog.ShowModal() == wx.ID_OK:
            method = dialog.GetStringSelection()

            if (
                method
                == "Subtract background (fast but requires static background & stable illumination)"
            ):
                self.use_detector = False

                contrasts = [
                    "Animal brighter than background",
                    "Animal darker than background",
                    "Hard to tell",
                ]
                dialog1 = wx.SingleChoiceDialog(
                    self,
                    message="Select the scenario that fits your videos best",
                    caption="Which fits best?",
                    choices=contrasts,
                )

                if dialog1.ShowModal() == wx.ID_OK:
                    contrast = dialog1.GetStringSelection()
                    if contrast == "Animal brighter than background":
                        self.animal_vs_bg = 0
                    elif contrast == "Animal darker than background":
                        self.animal_vs_bg = 1
                    else:
                        self.animal_vs_bg = 2
                    dialog2 = wx.MessageDialog(
                        self,
                        'Load an existing background from a folder?\nSelect "No" if dont know what it is.',
                        "(Optional) load existing background?",
                        wx.YES_NO | wx.ICON_QUESTION,
                    )
                    if dialog2.ShowModal() == wx.ID_YES:
                        dialog3 = wx.DirDialog(
                            self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE
                        )
                        if dialog3.ShowModal() == wx.ID_OK:
                            self.background_path = dialog3.GetPath()
                        dialog3.Destroy()
                    else:
                        self.background_path = None
                        if self.animal_vs_bg != 2:
                            dialog3 = wx.MessageDialog(
                                self,
                                'Unstable illumination in the video?\nSelect "Yes" if dont know what it is.',
                                "(Optional) unstable illumination?",
                                wx.YES_NO | wx.ICON_QUESTION,
                            )
                            if dialog3.ShowModal() == wx.ID_YES:
                                self.stable_illumination = False
                            else:
                                self.stable_illumination = True
                            dialog3.Destroy()
                    dialog2.Destroy()

                    if self.background_path is None:
                        ex_methods = [
                            "Use the entire duration (default but NOT recommended)",
                            'Decode from filenames: "_xst_" and "_xet_"',
                            "Enter two time points",
                        ]
                        dialog2 = wx.SingleChoiceDialog(
                            self,
                            message="Specify the time window for background extraction",
                            caption="Time window for background extraction",
                            choices=ex_methods,
                        )
                        if dialog2.ShowModal() == wx.ID_OK:
                            ex_method = dialog2.GetStringSelection()
                            if (
                                ex_method
                                == "Use the entire duration (default but NOT recommended)"
                            ):
                                self.decode_extraction = False
                                if self.animal_vs_bg == 0:
                                    self.text_detection.SetLabel(
                                        "Background subtraction: animal brighter, using the entire duration."
                                    )
                                elif self.animal_vs_bg == 1:
                                    self.text_detection.SetLabel(
                                        "Background subtraction: animal darker, using the entire duration."
                                    )
                                else:
                                    self.text_detection.SetLabel(
                                        "Background subtraction: animal partially brighter/darker, using the entire duration."
                                    )
                            elif (
                                ex_method
                                == 'Decode from filenames: "_xst_" and "_xet_"'
                            ):
                                self.decode_extraction = True
                                if self.animal_vs_bg == 0:
                                    self.text_detection.SetLabel(
                                        'Background subtraction: animal brighter, using time window decoded from filenames "_xst_" and "_xet_".'
                                    )
                                elif self.animal_vs_bg == 1:
                                    self.text_detection.SetLabel(
                                        'Background subtraction: animal darker, using time window decoded from filenames "_xst_" and "_xet_".'
                                    )
                                else:
                                    self.text_detection.SetLabel(
                                        'Background subtraction: animal partially brighter/darker, using time window decoded from filenames "_xst_" and "_xet_".'
                                    )
                            else:
                                self.decode_extraction = False
                                dialog3 = wx.NumberEntryDialog(
                                    self,
                                    "Enter the start time",
                                    "The unit is second:",
                                    "Start time for background extraction",
                                    0,
                                    0,
                                    100000000000000,
                                )
                                if dialog3.ShowModal() == wx.ID_OK:
                                    self.ex_start = int(dialog3.GetValue())
                                dialog3.Destroy()
                                dialog3 = wx.NumberEntryDialog(
                                    self,
                                    "Enter the end time",
                                    "The unit is second:",
                                    "End time for background extraction",
                                    0,
                                    0,
                                    100000000000000,
                                )
                                if dialog3.ShowModal() == wx.ID_OK:
                                    self.ex_end = int(dialog3.GetValue())
                                    if self.ex_end == 0:
                                        self.ex_end = None
                                dialog3.Destroy()
                                if self.animal_vs_bg == 0:
                                    if self.ex_end is None:
                                        self.text_detection.SetLabel(
                                            "Background subtraction: animal brighter, using time window (in seconds) from "
                                            + str(self.ex_start)
                                            + " to the end."
                                        )
                                    else:
                                        self.text_detection.SetLabel(
                                            "Background subtraction: animal brighter, using time window (in seconds) from "
                                            + str(self.ex_start)
                                            + " to "
                                            + str(self.ex_end)
                                            + "."
                                        )
                                elif self.animal_vs_bg == 1:
                                    if self.ex_end is None:
                                        self.text_detection.SetLabel(
                                            "Background subtraction: animal darker, using time window (in seconds) from "
                                            + str(self.ex_start)
                                            + " to the end."
                                        )
                                    else:
                                        self.text_detection.SetLabel(
                                            "Background subtraction: animal darker, using time window (in seconds) from "
                                            + str(self.ex_start)
                                            + " to "
                                            + str(self.ex_end)
                                            + "."
                                        )
                                else:
                                    if self.ex_end is None:
                                        self.text_detection.SetLabel(
                                            "Background subtraction: animal partially brighter/darker, using time window (in seconds) from "
                                            + str(self.ex_start)
                                            + " to the end."
                                        )
                                    else:
                                        self.text_detection.SetLabel(
                                            "Background subtraction: animal partially brighter/darker, using time window (in seconds) from "
                                            + str(self.ex_start)
                                            + " to "
                                            + str(self.ex_end)
                                            + "."
                                        )
                        dialog2.Destroy()

                dialog1.Destroy()

            else:
                self.use_detector = True
                self.animal_number = {}
                self.detector_path = os.path.join(
                    the_absolute_current_path, "detectors"
                )

                detectors = [
                    i
                    for i in os.listdir(self.detector_path)
                    if os.path.isdir(os.path.join(self.detector_path, i))
                ]
                if "__pycache__" in detectors:
                    detectors.remove("__pycache__")
                if "__init__" in detectors:
                    detectors.remove("__init__")
                if "__init__.py" in detectors:
                    detectors.remove("__init__.py")
                detectors.sort()
                if "Choose a new directory of the Detector" not in detectors:
                    detectors.append("Choose a new directory of the Detector")

                dialog1 = wx.SingleChoiceDialog(
                    self,
                    message="Select a Detector for animal detection",
                    caption="Select a Detector",
                    choices=detectors,
                )
                if dialog1.ShowModal() == wx.ID_OK:
                    detector = dialog1.GetStringSelection()
                    if detector == "Choose a new directory of the Detector":
                        dialog2 = wx.DirDialog(
                            self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE
                        )
                        if dialog2.ShowModal() == wx.ID_OK:
                            self.path_to_detector = dialog2.GetPaths()
                        dialog2.Destroy()
                    else:
                        self.path_to_detector = os.path.join(
                            self.detector_path, detector
                        )
                    with open(
                        os.path.join(self.path_to_detector, "model_parameters.txt")
                    ) as f:
                        model_parameters = f.read()
                    animal_names = json.loads(model_parameters)["animal_names"]
                    if len(animal_names) > 1:
                        dialog2 = wx.MultiChoiceDialog(
                            self,
                            message="Specify which animals/objects involved in analysis",
                            caption="Animal/Object kind",
                            choices=animal_names,
                        )
                        if dialog2.ShowModal() == wx.ID_OK:
                            self.animal_kinds = [
                                animal_names[i] for i in dialog2.GetSelections()
                            ]
                        else:
                            self.animal_kinds = animal_names
                        dialog2.Destroy()
                    else:
                        self.animal_kinds = animal_names
                    if self.behavior_mode >= 3:
                        dialog2 = wx.NumberEntryDialog(
                            self,
                            "Enter the Detector's detection threshold (0~100%)",
                            "The higher detection threshold,\nthe higher detection accuracy,\nbut the lower detection sensitivity.\nEnter 0 if don't know how to set.",
                            "Detection threshold",
                            0,
                            0,
                            100,
                        )
                        if dialog2.ShowModal() == wx.ID_OK:
                            detection_threshold = dialog2.GetValue()
                            self.detection_threshold = detection_threshold / 100
                        self.text_detection.SetLabel(
                            "Detector: "
                            + detector
                            + " (detection threshold: "
                            + str(detection_threshold)
                            + "%); The animals/objects: "
                            + str(self.animal_kinds)
                            + "."
                        )
                        dialog2.Destroy()
                    else:
                        for animal_name in self.animal_kinds:
                            self.animal_number[animal_name] = 1
                        self.text_animalnumber.SetLabel(
                            "The number of "
                            + str(self.animal_kinds)
                            + " is: "
                            + str(list(self.animal_number.values()))
                            + "."
                        )
                        self.text_detection.SetLabel(
                            "Detector: "
                            + detector
                            + "; "
                            + "The animals/objects: "
                            + str(self.animal_kinds)
                            + "."
                        )
                dialog1.Destroy()

                if self.behavior_mode < 3:
                    if torch.cuda.is_available():
                        dialog1 = wx.NumberEntryDialog(
                            self,
                            "Enter the batch size for faster processing",
                            "GPU is available in this device for Detectors.\nYou may use batch processing for faster speed.",
                            "Batch size",
                            1,
                            1,
                            500,
                        )
                        if dialog1.ShowModal() == wx.ID_OK:
                            self.detector_batch = int(dialog1.GetValue())
                        else:
                            self.detector_batch = 1
                        dialog1.Destroy()

        dialog.Destroy()

    @property
    def detector_batch(self):
        """The batch size to use when processing videos with a Detector."""
        return self._detector_batch

    @detector_batch.setter
    def detector_batch(self, detector_batch: int):
        """The batch size to use when processing videos with a Detector."""
        if detector_batch <= 0:
            raise ValueError("Batch size must be greater than or equal to 1.")
        self._detector_batch = detector_batch

    def specify_timing(self, event):
        if self.behavior_mode >= 3:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )

        else:
            if self.use_detector is False:
                dialog = wx.MessageDialog(
                    self,
                    "light on and off in videos?",
                    "Illumination shifts?",
                    wx.YES_NO | wx.ICON_QUESTION,
                )
                if dialog.ShowModal() == wx.ID_YES:
                    self.delta = 1.2
                else:
                    self.delta = 10000
                dialog.Destroy()

            if self.delta == 1.2 and self.use_detector is False:
                methods = [
                    "Automatic (for light on and off)",
                    'Decode from filenames: "_bt_"',
                    "Enter a time point",
                ]
            else:
                methods = ['Decode from filenames: "_bt_"', "Enter a time point"]

            dialog = wx.SingleChoiceDialog(
                self,
                message="Specify beginning time of analysis",
                caption="Beginning time of analysis",
                choices=methods,
            )
            if dialog.ShowModal() == wx.ID_OK:
                method = dialog.GetStringSelection()
                if method == "Automatic (for light on and off)":
                    self.autofind_t = True
                    self.decode_t = False
                    self.text_startanalyze.SetLabel(
                        "Automatically find the onset of the 1st time when light on / off as the beginning time."
                    )
                elif method == 'Decode from filenames: "_bt_"':
                    self.autofind_t = False
                    self.decode_t = True
                    self.text_startanalyze.SetLabel(
                        'Decode the beginning time from the filenames: the "t" immediately after the letter "b"" in "_bt_".'
                    )
                else:
                    self.autofind_t = False
                    self.decode_t = False
                    dialog2 = wx.NumberEntryDialog(
                        self,
                        "Enter the beginning time of analysis",
                        "The unit is second:",
                        "Beginning time of analysis",
                        0,
                        0,
                        100000000000000,
                    )
                    if dialog2.ShowModal() == wx.ID_OK:
                        self.t = float(dialog2.GetValue())
                        if self.t < 0:
                            self.t = 0
                        self.text_startanalyze.SetLabel(
                            "Analysis will begin at the: " + str(self.t) + " second."
                        )
                    dialog2.Destroy()
            dialog.Destroy()

    def input_duration(self, event):
        if self.behavior_mode >= 3:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )

        else:
            dialog = wx.NumberEntryDialog(
                self,
                "Enter the duration of the analysis",
                "The unit is second:",
                "Analysis duration",
                0,
                0,
                100000000000000,
            )
            if dialog.ShowModal() == wx.ID_OK:
                self.duration = int(dialog.GetValue())
                if self.duration != 0:
                    self.text_duration.SetLabel(
                        "The analysis duration is " + str(self.duration) + " seconds."
                    )
                else:
                    self.text_duration.SetLabel(
                        "The analysis duration is from the specified beginning time to the end of a video."
                    )
            dialog.Destroy()

    def specify_animalnumber(self, event):
        if self.behavior_mode >= 3:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )

        else:
            methods = ['Decode from filenames: "_nn_"', "Enter the number of animals"]

            dialog = wx.SingleChoiceDialog(
                self,
                message="Specify the number of animals in a video",
                caption="The number of animals in a video",
                choices=methods,
            )
            if dialog.ShowModal() == wx.ID_OK:
                method = dialog.GetStringSelection()
                if method == "Enter the number of animals":
                    self.decode_animalnumber = False
                    if self.use_detector is True:
                        self.animal_number = {}
                        for animal_name in self.animal_kinds:
                            dialog1 = wx.NumberEntryDialog(
                                self,
                                "",
                                "The number of " + str(animal_name) + ": ",
                                str(animal_name) + " number",
                                1,
                                1,
                                100,
                            )
                            if dialog1.ShowModal() == wx.ID_OK:
                                self.animal_number[animal_name] = int(
                                    dialog1.GetValue()
                                )
                            else:
                                self.animal_number[animal_name] = 1
                            dialog1.Destroy()
                        self.text_animalnumber.SetLabel(
                            "The number of "
                            + str(self.animal_kinds)
                            + " is: "
                            + str(list(self.animal_number.values()))
                            + "."
                        )
                    else:
                        dialog1 = wx.NumberEntryDialog(
                            self,
                            "",
                            "The number of animals:",
                            "Animal number",
                            1,
                            1,
                            100,
                        )
                        if dialog1.ShowModal() == wx.ID_OK:
                            self.animal_number = int(dialog1.GetValue())
                        else:
                            self.animal_number = 1
                        self.text_animalnumber.SetLabel(
                            "The total number of animals in a video is "
                            + str(self.animal_number)
                            + "."
                        )
                        dialog1.Destroy()
                else:
                    self.decode_animalnumber = True
                    self.text_animalnumber.SetLabel(
                        'Decode from the filenames: the "n" immediately after the letter "n" in _"nn"_.'
                    )
            dialog.Destroy()

    def select_behaviors(self, event):
        if self.path_to_categorizer is None:
            wx.MessageBox(
                "No Categorizer selected! The behavior names are listed in the Categorizer.",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )

        else:
            if len(self.animal_kinds) > 1:
                dialog = wx.MultiChoiceDialog(
                    self,
                    message="Specify which animals/objects to annotate",
                    caption="Animal/Object to annotate",
                    choices=self.animal_kinds,
                )
                if dialog.ShowModal() == wx.ID_OK:
                    self.animal_to_include = [
                        self.animal_kinds[i] for i in dialog.GetSelections()
                    ]
                else:
                    self.animal_to_include = self.animal_kinds
                dialog.Destroy()
            else:
                self.animal_to_include = self.animal_kinds

            dialog = wx.MultiChoiceDialog(
                self,
                message="Select behaviors",
                caption="Behaviors to annotate",
                choices=list(self.behaviornames_and_colors.keys()),
            )
            if dialog.ShowModal() == wx.ID_OK:
                self.behavior_to_include = [
                    list(self.behaviornames_and_colors.keys())[i]
                    for i in dialog.GetSelections()
                ]
            else:
                self.behavior_to_include = list(self.behaviornames_and_colors.keys())
            dialog.Destroy()

            if len(self.behavior_to_include) == 0:
                self.behavior_to_include = list(self.behaviornames_and_colors.keys())
            if self.behavior_to_include[0] == "all":
                self.behavior_to_include = list(self.behaviornames_and_colors.keys())

            if self.behavior_mode == 2:
                dialog = wx.MessageDialog(
                    self,
                    "Specify individual-specific behaviors? e.g., sex-specific behaviors only occur in a specific sex and\ncan be used to maintain the correct ID of this individual during the entire analysis.",
                    "Individual-specific behaviors?",
                    wx.YES_NO | wx.ICON_QUESTION,
                )
                if dialog.ShowModal() == wx.ID_YES:
                    for animal_name in self.animal_kinds:
                        dialog1 = wx.MultiChoiceDialog(
                            self,
                            message="Select individual-specific behaviors for "
                            + str(animal_name),
                            caption="Individual-specific behaviors for "
                            + str(animal_name),
                            choices=self.behavior_to_include,
                        )
                        if dialog1.ShowModal() == wx.ID_OK:
                            self.specific_behaviors[animal_name] = {}
                            self.correct_ID = True
                            specific_behaviors = [
                                self.behavior_to_include[i]
                                for i in dialog1.GetSelections()
                            ]
                            for specific_behavior in specific_behaviors:
                                self.specific_behaviors[animal_name][
                                    specific_behavior
                                ] = None
                        dialog1.Destroy()
                else:
                    self.correct_ID = False
                dialog.Destroy()

            complete_colors = list(mpl.colors.cnames.values())
            colors = []
            for c in complete_colors:
                colors.append(["#ffffff", c])

            dialog = wx.MessageDialog(
                self,
                "Specify the color to represent\nthe behaviors in annotations and plots?",
                "Specify colors for behaviors?",
                wx.YES_NO | wx.ICON_QUESTION,
            )
            if dialog.ShowModal() == wx.ID_YES:
                names_colors = {}
                n = 0
                while n < len(self.behavior_to_include):
                    dialog2 = ColorPicker(
                        self,
                        "Color for " + self.behavior_to_include[n],
                        [self.behavior_to_include[n], colors[n]],
                    )
                    if dialog2.ShowModal() == wx.ID_OK:
                        (r, b, g, _) = dialog2.color_picker.GetColour()
                        new_color = "#%02x%02x%02x" % (r, b, g)
                        self.behaviornames_and_colors[self.behavior_to_include[n]] = [
                            "#ffffff",
                            new_color,
                        ]
                        names_colors[self.behavior_to_include[n]] = new_color
                    else:
                        if n < len(colors):
                            names_colors[self.behavior_to_include[n]] = colors[n][1]
                            self.behaviornames_and_colors[
                                self.behavior_to_include[n]
                            ] = colors[n]
                    dialog2.Destroy()
                    n += 1
                if self.correct_ID is True:
                    self.text_selectbehaviors.SetLabel(
                        "Selected: "
                        + str(list(names_colors.keys()))
                        + ". Specific behaviors: "
                        + str(self.specific_behaviors)
                        + "."
                    )
                else:
                    self.text_selectbehaviors.SetLabel(
                        "Selected: " + str(list(names_colors.keys())) + "."
                    )
            else:
                for color in colors:
                    index = colors.index(color)
                    if index < len(self.behavior_to_include):
                        behavior_name = list(self.behaviornames_and_colors.keys())[
                            index
                        ]
                        self.behaviornames_and_colors[behavior_name] = color
                if self.correct_ID is True:
                    self.text_selectbehaviors.SetLabel(
                        "Selected: "
                        + str(self.behavior_to_include)
                        + " with default colors. Specific behaviors:"
                        + str(self.specific_behaviors)
                        + "."
                    )
                else:
                    self.text_selectbehaviors.SetLabel(
                        "Selected: "
                        + str(self.behavior_to_include)
                        + " with default colors."
                    )
            dialog.Destroy()

            if self.behavior_mode != 3:
                dialog = wx.MessageDialog(
                    self,
                    "Show legend of behavior names in the annotated video?",
                    "Legend in video?",
                    wx.YES_NO | wx.ICON_QUESTION,
                )
                if dialog.ShowModal() == wx.ID_YES:
                    self.show_legend = True
                else:
                    self.show_legend = False
                dialog.Destroy()

    def select_parameters(self, event):
        if self.behavior_mode >= 3:
            wx.MessageBox(
                'No need to specify this since the selected behavior mode is "Static images".',
                "Error",
                wx.OK | wx.ICON_ERROR,
            )

        else:
            if self.path_to_categorizer is None:
                parameters = [
                    "3 areal parameters",
                    "3 length parameters",
                    "4 locomotion parameters",
                ]
            else:
                if self.behavior_mode == 1:
                    parameters = ["count", "duration", "latency"]
                else:
                    parameters = [
                        "count",
                        "duration",
                        "latency",
                        "3 areal parameters",
                        "3 length parameters",
                        "4 locomotion parameters",
                    ]

            dialog = wx.MultiChoiceDialog(
                self,
                message="Select quantitative measurements",
                caption="Quantitative measurements",
                choices=parameters,
            )
            if dialog.ShowModal() == wx.ID_OK:
                self.parameter_to_analyze = [
                    parameters[i] for i in dialog.GetSelections()
                ]
            else:
                self.parameter_to_analyze = []
            dialog.Destroy()

            if len(self.parameter_to_analyze) <= 0:
                self.parameter_to_analyze = []
                self.normalize_distance = False
                self.text_selectparameters.SetLabel("NO parameter selected.")
            else:
                if "4 locomotion parameters" in self.parameter_to_analyze:
                    dialog = wx.MessageDialog(
                        self,
                        "Normalize the distances by the size of an animal? If no, all distances will be output in pixels.",
                        "Normalize the distances?",
                        wx.YES_NO | wx.ICON_QUESTION,
                    )
                    if dialog.ShowModal() == wx.ID_YES:
                        self.normalize_distance = True
                        self.text_selectparameters.SetLabel(
                            "Selected: "
                            + str(self.parameter_to_analyze)
                            + "; with normalization of distance."
                        )
                    else:
                        self.normalize_distance = False
                        self.text_selectparameters.SetLabel(
                            "Selected: "
                            + str(self.parameter_to_analyze)
                            + "; NO normalization of distance."
                        )
                    dialog.Destroy()
                else:
                    self.normalize_distance = False
                    self.text_selectparameters.SetLabel(
                        "Selected: " + str(self.parameter_to_analyze) + "."
                    )

    def analyze_behaviors(self, event):
        if self.path_to_videos is None or self.result_path is None:
            wx.MessageBox(
                "No input video(s) / result folder.", "Error", wx.OK | wx.ICON_ERROR
            )

        else:
            if self.behavior_mode >= 3:
                if self.path_to_categorizer is None or self.path_to_detector is None:
                    wx.MessageBox(
                        "You need to select a Categorizer / Detector.",
                        "Error",
                        wx.OK | wx.ICON_ERROR,
                    )
                else:
                    AAD = AnalyzeAnimalDetector()
                    AAD.analyze_images_individuals(
                        self.path_to_detector,
                        self.path_to_videos,
                        self.result_path,
                        self.animal_kinds,
                        path_to_categorizer=self.path_to_categorizer,
                        generate=False,
                        animal_to_include=self.animal_to_include,
                        behavior_to_include=self.behavior_to_include,
                        names_and_colors=self.behaviornames_and_colors,
                        imagewidth=self.framewidth,
                        dim_conv=self.dim_conv,
                        channel=self.channel,
                        detection_threshold=self.detection_threshold,
                        uncertain=self.uncertain,
                        background_free=self.background_free,
                        social_distance=0,
                    )

            else:
                all_events = {}
                event_data = {}
                all_lengths = []

                if self.use_detector is True:
                    for animal_name in self.animal_kinds:
                        all_events[animal_name] = {}
                    if len(self.animal_to_include) == 0:
                        self.animal_to_include = self.animal_kinds

                if self.path_to_categorizer is None:
                    self.behavior_to_include = []
                else:
                    if self.behavior_to_include[0] == "all":
                        self.behavior_to_include = list(
                            self.behaviornames_and_colors.keys()
                        )

                for i in self.path_to_videos:
                    filename = os.path.splitext(os.path.basename(i))[0].split("_")
                    if self.decode_animalnumber is True:
                        if self.use_detector is True:
                            self.animal_number = {}
                            number = [
                                x[1:] for x in filename if len(x) > 1 and x[0] == "n"
                            ]
                            for a, animal_name in enumerate(self.animal_kinds):
                                self.animal_number[animal_name] = int(number[a])
                        else:
                            for x in filename:
                                if len(x) > 1:
                                    if x[0] == "n":
                                        self.animal_number = int(x[1:])
                    if self.decode_t is True:
                        for x in filename:
                            if len(x) > 1:
                                if x[0] == "b":
                                    self.t = float(x[1:])
                    if self.decode_extraction is True:
                        for x in filename:
                            if len(x) > 2:
                                if x[:2] == "xs":
                                    self.ex_start = int(x[2:])
                                if x[:2] == "xe":
                                    self.ex_end = int(x[2:])

                    if self.animal_number is None:
                        if self.use_detector is True:
                            self.animal_number = {}
                            for animal_name in self.animal_kinds:
                                self.animal_number[animal_name] = 1
                        else:
                            self.animal_number = 1

                    if self.path_to_categorizer is None:
                        self.behavior_mode = 0
                        categorize_behavior = False
                    else:
                        categorize_behavior = True

                    if self.use_detector is False:
                        AA = AnalyzeAnimal()
                        AA.prepare_analysis(
                            i,
                            self.result_path,
                            self.animal_number,
                            delta=self.delta,
                            names_and_colors=self.behaviornames_and_colors,
                            framewidth=self.framewidth,
                            stable_illumination=self.stable_illumination,
                            dim_tconv=self.dim_tconv,
                            dim_conv=self.dim_conv,
                            channel=self.channel,
                            include_bodyparts=self.include_bodyparts,
                            std=self.std,
                            categorize_behavior=categorize_behavior,
                            animation_analyzer=self.animation_analyzer,
                            path_background=self.background_path,
                            autofind_t=self.autofind_t,
                            t=self.t,
                            duration=self.duration,
                            ex_start=self.ex_start,
                            ex_end=self.ex_end,
                            length=self.length,
                            animal_vs_bg=self.animal_vs_bg,
                        )
                        if self.behavior_mode == 0:
                            AA.acquire_information(background_free=self.background_free)
                            AA.craft_data()
                            interact_all = False
                        else:
                            AA.acquire_information_interact_basic(
                                background_free=self.background_free
                            )
                            interact_all = True
                        if self.path_to_categorizer is not None:
                            AA.categorize_behaviors(
                                self.path_to_categorizer, uncertain=self.uncertain
                            )
                        AA.annotate_video(
                            self.behavior_to_include,
                            show_legend=self.show_legend,
                            interact_all=interact_all,
                        )
                        AA.export_results(
                            normalize_distance=self.normalize_distance,
                            parameter_to_analyze=self.parameter_to_analyze,
                        )

                        if self.path_to_categorizer is not None:
                            for n in AA.event_probability:
                                all_events[len(all_events)] = AA.event_probability[n]
                                all_lengths.append(len(AA.event_probability[n]))

                    else:
                        AAD = AnalyzeAnimalDetector()
                        AAD.prepare_analysis(
                            self.path_to_detector,
                            i,
                            self.result_path,
                            self.animal_number,
                            self.animal_kinds,
                            self.behavior_mode,
                            names_and_colors=self.behaviornames_and_colors,
                            framewidth=self.framewidth,
                            dim_tconv=self.dim_tconv,
                            dim_conv=self.dim_conv,
                            channel=self.channel,
                            include_bodyparts=self.include_bodyparts,
                            std=self.std,
                            categorize_behavior=categorize_behavior,
                            animation_analyzer=self.animation_analyzer,
                            t=self.t,
                            duration=self.duration,
                            length=self.length,
                            social_distance=self.social_distance,
                        )
                        AAD.acquire_information(
                            batch_size=self.detector_batch,
                            background_free=self.background_free,
                        )
                        if self.behavior_mode != 1:
                            AAD.craft_data()
                        if self.path_to_categorizer is not None:
                            AAD.categorize_behaviors(
                                self.path_to_categorizer, uncertain=self.uncertain
                            )
                        if self.correct_ID is True:
                            AAD.correct_identity(self.specific_behaviors)
                        AAD.annotate_video(
                            self.animal_to_include,
                            self.behavior_to_include,
                            show_legend=self.show_legend,
                        )
                        AAD.export_results(
                            normalize_distance=self.normalize_distance,
                            parameter_to_analyze=self.parameter_to_analyze,
                        )

                        if self.path_to_categorizer is not None:
                            for animal_name in self.animal_kinds:
                                for n in AAD.event_probability[animal_name]:
                                    all_events[animal_name][
                                        len(all_events[animal_name])
                                    ] = AAD.event_probability[animal_name][n]
                                    all_lengths.append(
                                        len(AAD.event_probability[animal_name][n])
                                    )

                if self.path_to_categorizer is not None:
                    if self.use_detector is False:
                        for n in all_events:
                            event_data[len(event_data)] = all_events[n][
                                : min(all_lengths)
                            ]
                        time_points = AA.all_time[: min(all_lengths)]
                        all_events_df = pd.DataFrame(event_data, index=time_points)
                        all_events_df.to_excel(
                            os.path.join(self.result_path, "all_events.xlsx"),
                            float_format="%.2f",
                            index_label="time/ID",
                        )
                        plot_evnets(
                            self.result_path,
                            event_data,
                            time_points,
                            self.behaviornames_and_colors,
                            self.behavior_to_include,
                            width=0,
                            height=0,
                        )
                        folders = [
                            i
                            for i in os.listdir(self.result_path)
                            if os.path.isdir(os.path.join(self.result_path, i))
                        ]
                        folders.sort()
                        for behavior_name in self.behaviornames_and_colors:
                            all_summary = []
                            for folder in folders:
                                individual_summary = os.path.join(
                                    self.result_path,
                                    folder,
                                    behavior_name,
                                    "all_summary.xlsx",
                                )
                                if os.path.exists(individual_summary) is True:
                                    all_summary.append(
                                        pd.read_excel(individual_summary)
                                    )
                            if len(all_summary) >= 1:
                                all_summary = pd.concat(all_summary, ignore_index=True)
                                all_summary.to_excel(
                                    os.path.join(
                                        self.result_path,
                                        behavior_name + "_summary.xlsx",
                                    ),
                                    float_format="%.2f",
                                    index_label="ID/parameter",
                                )

                    else:
                        for animal_name in self.animal_to_include:
                            for n in all_events[animal_name]:
                                event_data[len(event_data)] = all_events[animal_name][
                                    n
                                ][: min(all_lengths)]
                            event_data[len(event_data)] = [["NA", -1]] * min(
                                all_lengths
                            )
                        del event_data[len(event_data) - 1]
                        time_points = AAD.all_time[: min(all_lengths)]
                        all_events_df = pd.DataFrame(event_data, index=time_points)
                        all_events_df.to_excel(
                            os.path.join(self.result_path, "all_events.xlsx"),
                            float_format="%.2f",
                            index_label="time/ID",
                        )
                        plot_evnets(
                            self.result_path,
                            event_data,
                            time_points,
                            self.behaviornames_and_colors,
                            self.behavior_to_include,
                            width=0,
                            height=0,
                        )
                        folders = [
                            i
                            for i in os.listdir(self.result_path)
                            if os.path.isdir(os.path.join(self.result_path, i))
                        ]
                        folders.sort()
                        for animal_name in self.animal_kinds:
                            for behavior_name in self.behaviornames_and_colors:
                                all_summary = []
                                for folder in folders:
                                    individual_summary = os.path.join(
                                        self.result_path,
                                        folder,
                                        behavior_name,
                                        animal_name + "_all_summary.xlsx",
                                    )
                                    if os.path.exists(individual_summary) is True:
                                        all_summary.append(
                                            pd.read_excel(individual_summary)
                                        )
                                if len(all_summary) >= 1:
                                    all_summary = pd.concat(
                                        all_summary, ignore_index=True
                                    )
                                    all_summary.to_excel(
                                        os.path.join(
                                            self.result_path,
                                            animal_name
                                            + "_"
                                            + behavior_name
                                            + "_summary.xlsx",
                                        ),
                                        float_format="%.2f",
                                        index_label="ID/parameter",
                                    )



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""


import matplotlib

from .tools import *

matplotlib.use("Agg")
import datetime
import os
import random
import shutil
from collections import deque
from pathlib import Path

import cv2
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.ndimage as ndimage
import tensorflow as tf
from skimage import exposure, transform
from skimage.transform import AffineTransform
from sklearn.metrics import classification_report
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelBinarizer
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau
from tensorflow.keras.layers import (
    LSTM,
    Activation,
    Add,
    AveragePooling2D,
    BatchNormalization,
    Conv2D,
    Dense,
    Dropout,
    Flatten,
    Input,
    MaxPooling2D,
    TimeDistributed,
    ZeroPadding2D,
    concatenate,
)
from tensorflow.keras.models import Model, load_model
from tensorflow.keras.optimizers import SGD
from tensorflow.keras.preprocessing.image import img_to_array

CATEGORIZER_FOLDER = Path(__file__).resolve().parent / "models"


def get_categorizer_names() -> list[str]:
    """Return the names of all saved categorizers."""
    ignore = ["__pycache__", "__init__", "__init.py__"]
    return [
        path.name
        for path in CATEGORIZER_FOLDER.glob("*")
        if path.is_dir() and path.name not in ignore
    ]


def delete_categorizer(categorizer_name: str):
    """Permanently delete the given categorizer."""
    shutil.rmtree(str(CATEGORIZER_FOLDER / categorizer_name))


class Categorizers:
    def __init__(self):
        self.extension_image = (
            ".png",
            ".PNG",
            ".jpeg",
            ".JPEG",
            ".jpg",
            ".JPG",
            ".tiff",
            ".TIFF",
            ".bmp",
            ".BMP",
        )
        self.extension_video = (".avi", ".mpg", ".wmv", ".mp4", ".mkv", ".m4v", ".mov")
        self.classnames = None

    def rename_label(self, file_path, new_path, resize=None):
        folder_list = [
            i
            for i in os.listdir(file_path)
            if os.path.isdir(os.path.join(file_path, i))
        ]
        print("Behavior names are: " + str(folder_list))
        previous_lenth = None
        imagedata = False

        for folder in folder_list:
            name_list = [
                i
                for i in os.listdir(os.path.join(file_path, folder))
                if i.endswith(".avi")
            ]

            if len(name_list) == 0:
                name_list = [
                    i
                    for i in os.listdir(os.path.join(file_path, folder))
                    if i.endswith(".jpg")
                ]
                imagedata = True

            for i in name_list:
                if imagedata is True:
                    image = os.path.join(file_path, folder, i)
                    new_image = os.path.join(
                        new_path, str(name_list.index(i)) + "_" + folder + ".jpg"
                    )
                    image = cv2.imread(image)
                    if resize is not None:
                        image = cv2.resize(
                            image, (resize, resize), interpolation=cv2.INTER_AREA
                        )
                    cv2.imwrite(new_image, image)

                else:
                    animation = os.path.join(file_path, folder, i)
                    pattern_image = os.path.join(
                        file_path, folder, os.path.splitext(i)[0] + ".jpg"
                    )
                    current_length = 0

                    new_animation = os.path.join(
                        new_path, str(name_list.index(i)) + "_" + folder + ".avi"
                    )
                    new_pattern_image = os.path.join(
                        new_path, str(name_list.index(i)) + "_" + folder + ".jpg"
                    )
                    writer = None
                    capture = cv2.VideoCapture(animation)
                    fps = round(capture.get(cv2.CAP_PROP_FPS))
                    while True:
                        retval, frame = capture.read()
                        current_length += 1
                        if frame is None:
                            break
                        if resize is not None:
                            frame = cv2.resize(
                                frame, (resize, resize), interpolation=cv2.INTER_AREA
                            )
                        if writer is None:
                            (h, w) = frame.shape[:2]
                            writer = cv2.VideoWriter(
                                new_animation,
                                cv2.VideoWriter_fourcc(*"MJPG"),
                                fps,
                                (w, h),
                                True,
                            )
                        writer.write(frame)
                    capture.release()
                    writer.release()
                    pattern_image = cv2.imread(pattern_image)
                    if resize is not None:
                        pattern_image = cv2.resize(
                            pattern_image,
                            (resize, resize),
                            interpolation=cv2.INTER_AREA,
                        )
                    cv2.imwrite(new_pattern_image, pattern_image)
                    if previous_lenth is None:
                        previous_lenth = current_length
                    else:
                        if previous_lenth != current_length:
                            previous_lenth = current_length
                            print(
                                "Inconsistent duration of animation detected at: "
                                + str(i)
                                + ". Check the duration of animations!"
                            )

        print("All prepared training examples stored in: " + str(new_path))

    def build_data(
        self,
        path_to_animations,
        dim_tconv=0,
        dim_conv=64,
        channel=1,
        time_step=15,
        aug_methods=[],
        background_free=True,
        behavior_mode=0,
    ):
        animations = deque()
        pattern_images = deque()
        labels = deque()
        amount = 0

        if len(aug_methods) == 0:
            methods = ["orig"]

        else:
            remove = []

            all_methods = [
                "orig",
                "del1",
                "del2",
                "rot1",
                "rot2",
                "rot3",
                "rot4",
                "flph",
                "flpv",
                "brih",
                "bril",
                "shrp",
                "shrn",
                "sclh",
                "sclw",
                "rot1_brih",
                "rot1_bril",
                "rot4_brih",
                "rot4_bril",
                "flph_brih",
                "flph_bril",
                "flpv_brih",
                "flpv_bril",
                "del2_rot5_brii",
                "del2_rot6_brii",
                "del2_rot5_brid",
                "del2_rot6_brid",
                "del2_flph_brii",
                "del2_flph_brid",
                "del2_flpv_brii",
                "del2_flpv_brid",
                "flph_rot5_brii",
                "flph_rot6_brii",
                "flph_rot5_brid",
                "flph_rot6_brid",
                "flpv_rot5_brii",
                "flpv_rot6_brii",
                "flpv_rot5_brid",
                "flpv_rot6_brid",
                "flph_shrp_brii",
                "flph_shrn_brii",
                "flph_shrp_brid",
                "flph_shrn_brid",
                "flpv_shrp_brii",
                "flpv_shrn_brii",
                "flpv_shrp_brid",
                "flpv_shrn_brid",
            ]

            for i in all_methods:
                if "random rotation" not in aug_methods:
                    if "rot" in i:
                        remove.append(i)
                if "horizontal flipping" not in aug_methods:
                    if "flph" in i:
                        remove.append(i)
                if "vertical flipping" not in aug_methods:
                    if "flpv" in i:
                        remove.append(i)
                if "random brightening" not in aug_methods:
                    if "brih" in i:
                        remove.append(i)
                if "random dimming" not in aug_methods:
                    if "bril" in i:
                        remove.append(i)
                if "random shearing" not in aug_methods:
                    if "shr" in i:
                        remove.append(i)
                if "random rescaling" not in aug_methods:
                    if "scl" in i:
                        remove.append(i)
                if "random deletion" not in aug_methods:
                    if "del" in i:
                        remove.append(i)
                if "exclude original" in aug_methods:
                    remove.append("orig")

            methods = list(set(all_methods) - set(remove))

        for i in path_to_animations:
            name = os.path.splitext(os.path.basename(i))[0].split("_")[0]
            label = os.path.splitext(i)[0].split("_")[-1]
            path_to_pattern_image = os.path.splitext(i)[0] + ".jpg"

            random.shuffle(methods)

            for m in methods:
                if "rot1" in m:
                    angle = np.random.uniform(10, 50)
                elif "rot2" in m:
                    angle = np.random.uniform(50, 90)
                elif "rot3" in m:
                    angle = np.random.uniform(90, 130)
                elif "rot4" in m:
                    angle = np.random.uniform(130, 170)
                elif "rot5" in m:
                    angle = np.random.uniform(30, 80)
                elif "rot6" in m:
                    angle = np.random.uniform(100, 150)
                else:
                    angle = None

                if "flph" in m:
                    code = 1
                elif "flpv" in m:
                    code = 0
                else:
                    code = None

                if "brih" in m:
                    beta = np.random.uniform(30, 80)
                elif "brii" in m:
                    beta = np.random.uniform(10, 45)
                elif "bril" in m:
                    beta = np.random.uniform(-80, -30)
                elif "brid" in m:
                    beta = np.random.uniform(-45, 10)
                else:
                    beta = None

                if "shrp" in m:
                    shear = np.random.uniform(0.15, 0.21)
                elif "shrn" in m:
                    shear = np.random.uniform(-0.21, -0.15)
                else:
                    shear = None

                if "sclh" in m:
                    width = 0
                    scale = np.random.uniform(0.6, 0.9)
                elif "sclw" in m:
                    width = 1
                    scale = np.random.uniform(0.6, 0.9)
                else:
                    scale = None

                if "del1" in m:
                    if time_step >= 30:
                        idx1 = random.randint(0, round(time_step / 3))
                        idx2 = random.randint(
                            round(time_step / 3) + 1, round(time_step * 2 / 3)
                        )
                        to_delete = [idx1, idx2]
                    else:
                        to_delete = [random.randint(0, round(time_step / 3))]
                elif "del2" in m:
                    to_delete = [random.randint(0, round(time_step / 2) + 1)]
                else:
                    to_delete = None

                if dim_tconv != 0:
                    capture = cv2.VideoCapture(i)
                    animation = deque()
                    frames = deque(maxlen=time_step)
                    original_frame = None
                    n = 0

                    while True:
                        retval, frame = capture.read()
                        if original_frame is None:
                            original_frame = frame
                        if frame is None:
                            break
                        frames.append(frame)

                    frames_length = len(frames)
                    if frames_length < time_step:
                        for diff in range(time_step - frames_length):
                            frames.append(np.zeros_like(original_frame))
                        print(
                            "Inconsistent duration of animation detected at: "
                            + str(i)
                            + "."
                        )
                        print(
                            "Zero padding has been used, which may decrease the training accuracy."
                        )

                    for frame in frames:
                        if to_delete is not None and n in to_delete:
                            blob = np.zeros_like(original_frame)

                        else:
                            frame_contrast = np.uint8(
                                exposure.rescale_intensity(frame, out_range=(0, 255))
                            )

                            if background_free is True:
                                frame_gray = cv2.cvtColor(
                                    frame_contrast, cv2.COLOR_BGR2GRAY
                                )
                                thred = cv2.threshold(
                                    frame_gray,
                                    0,
                                    255,
                                    cv2.THRESH_BINARY + cv2.THRESH_OTSU,
                                )[1]
                                cnts, _ = cv2.findContours(
                                    thred, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE
                                )
                                if len(cnts) == 0:
                                    blob = np.zeros_like(frame)
                                else:
                                    if behavior_mode == 0:
                                        contour = sorted(
                                            cnts, key=cv2.contourArea, reverse=True
                                        )[0]
                                        blob = extract_blob(frame, contour, channel=3)
                                    else:
                                        (y_bt, y_tp, x_lf, x_rt) = crop_frame(
                                            frame, cnts
                                        )
                                        blob = frame_contrast[y_bt:y_tp, x_lf:x_rt]
                            else:
                                blob = frame_contrast

                            if code is not None:
                                blob = cv2.flip(blob, code)

                            if beta is not None:
                                blob = blob.astype("float")
                                if background_free is True:
                                    blob[blob > 30] += beta
                                else:
                                    blob += beta
                                blob = np.uint8(np.clip(blob, 0, 255))

                            if angle is not None:
                                blob = ndimage.rotate(
                                    blob, angle, reshape=False, prefilter=False
                                )

                            if shear is not None:
                                tf = AffineTransform(shear=shear)
                                blob = transform.warp(
                                    blob,
                                    tf,
                                    order=1,
                                    preserve_range=True,
                                    mode="constant",
                                )

                            if scale is not None:
                                blob_black = np.zeros_like(blob)
                                if width == 0:
                                    blob_scl = cv2.resize(
                                        blob,
                                        (blob.shape[1], int(blob.shape[0] * scale)),
                                        interpolation=cv2.INTER_AREA,
                                    )
                                else:
                                    blob_scl = cv2.resize(
                                        blob,
                                        (int(blob.shape[1] * scale), blob.shape[0]),
                                        interpolation=cv2.INTER_AREA,
                                    )
                                blob_scl = img_to_array(blob_scl)
                                x = (blob_black.shape[1] - blob_scl.shape[1]) // 2
                                y = (blob_black.shape[0] - blob_scl.shape[0]) // 2
                                blob_black[
                                    y : y + blob_scl.shape[0], x : x + blob_scl.shape[1]
                                ] = blob_scl
                                blob = blob_black

                        if channel == 1:
                            blob = cv2.cvtColor(np.uint8(blob), cv2.COLOR_BGR2GRAY)

                        blob = cv2.resize(
                            blob, (dim_tconv, dim_tconv), interpolation=cv2.INTER_AREA
                        )
                        blob = img_to_array(blob)
                        animation.append(blob)

                        n += 1

                    capture.release()

                    animations.append(np.array(animation))

                pattern_image = cv2.imread(path_to_pattern_image)

                if code is not None:
                    pattern_image = cv2.flip(pattern_image, code)

                if behavior_mode >= 3:
                    if beta is not None:
                        if background_free is True:
                            pattern_image_gray = cv2.cvtColor(
                                pattern_image, cv2.COLOR_BGR2GRAY
                            )
                            thred = cv2.threshold(
                                pattern_image_gray,
                                0,
                                255,
                                cv2.THRESH_BINARY + cv2.THRESH_OTSU,
                            )[1]
                            cnts, _ = cv2.findContours(
                                thred, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE
                            )
                            if len(cnts) == 0:
                                pattern_image = np.zeros_like(pattern_image)
                            else:
                                contour = sorted(
                                    cnts, key=cv2.contourArea, reverse=True
                                )[0]
                                pattern_image = extract_blob(
                                    pattern_image, contour, channel=3
                                )
                            pattern_image = pattern_image.astype("float")
                            pattern_image[pattern_image > 30] += beta
                        else:
                            pattern_image += beta
                        pattern_image = np.uint8(np.clip(pattern_image, 0, 255))

                if angle is not None:
                    pattern_image = ndimage.rotate(
                        pattern_image, angle, reshape=False, prefilter=False
                    )

                if shear is not None:
                    tf = AffineTransform(shear=shear)
                    pattern_image = transform.warp(
                        pattern_image, tf, order=1, preserve_range=True, mode="constant"
                    )

                if scale is not None:
                    pattern_image_black = np.zeros_like(pattern_image)
                    if width == 0:
                        pattern_image_scl = cv2.resize(
                            pattern_image,
                            (
                                pattern_image.shape[1],
                                int(pattern_image.shape[0] * scale),
                            ),
                            interpolation=cv2.INTER_AREA,
                        )
                    else:
                        pattern_image_scl = cv2.resize(
                            pattern_image,
                            (
                                int(pattern_image.shape[1] * scale),
                                pattern_image.shape[0],
                            ),
                            interpolation=cv2.INTER_AREA,
                        )
                    x = (pattern_image_black.shape[1] - pattern_image_scl.shape[1]) // 2
                    y = (pattern_image_black.shape[0] - pattern_image_scl.shape[0]) // 2
                    pattern_image_black[
                        y : y + pattern_image_scl.shape[0],
                        x : x + pattern_image_scl.shape[1],
                        :,
                    ] = pattern_image_scl
                    pattern_image = pattern_image_black

                if behavior_mode >= 3:
                    if channel == 1:
                        pattern_image = cv2.cvtColor(
                            np.uint8(pattern_image), cv2.COLOR_BGR2GRAY
                        )

                pattern_image = cv2.resize(
                    pattern_image, (dim_conv, dim_conv), interpolation=cv2.INTER_AREA
                )
                pattern_images.append(img_to_array(pattern_image))

                labels.append(label)

                amount += 1
                if amount % 10000 == 0:
                    print("The augmented example amount: " + str(amount))
                    print(datetime.datetime.now())

        if dim_tconv != 0:
            animations = np.array(animations, dtype="float32") / 255.0
        pattern_images = np.array(pattern_images, dtype="float32") / 255.0
        labels = np.array(labels)

        return animations, pattern_images, labels

    def simple_vgg(self, inputs, filters, classes=3, level=2, with_classifier=False):
        if level < 2:
            layers = [2]
        elif level < 3:
            layers = [2, 3]
        elif level < 4:
            layers = [2, 3, 4]
        else:
            layers = [2, 3, 4, 4]

        for i in layers:
            for n in range(i):
                if n == 0:
                    if layers.index(i) == 0:
                        x = Conv2D(
                            filters,
                            kernel_size=(3, 3),
                            padding="same",
                            activation="relu",
                        )(inputs)
                        x = BatchNormalization()(x)
                    else:
                        x = Conv2D(
                            filters,
                            kernel_size=(3, 3),
                            padding="same",
                            activation="relu",
                        )(x)
                        x = BatchNormalization()(x)
                else:
                    x = Conv2D(
                        filters, kernel_size=(3, 3), padding="same", activation="relu"
                    )(x)
                    x = BatchNormalization()(x)
            x = MaxPooling2D(pool_size=(2, 2))(x)
            filters = int(filters * 2)

        x = Flatten()(x)

        if with_classifier is False:
            return x

        else:
            x = Dense(int(filters / 2), activation="relu")(x)
            x = BatchNormalization()(x)
            x = Dropout(0.5)(x)
            if classes == 2:
                x = Dense(1, activation="sigmoid")(x)
            else:
                x = Dense(classes, activation="softmax")(x)

            model = Model(inputs=inputs, outputs=x)
            # plot_model(model,'model.png',show_shapes=True)

            return model

    def simple_tvgg(self, inputs, filters, classes=3, level=2, with_classifier=False):
        if level < 2:
            layers = [2]
        elif level < 3:
            layers = [2, 3]
        elif level < 4:
            layers = [2, 3, 4]
        else:
            layers = [2, 3, 4, 4]

        for i in layers:
            for n in range(i):
                if n == 0:
                    if layers.index(i) == 0:
                        x = TimeDistributed(
                            Conv2D(
                                filters,
                                kernel_size=(3, 3),
                                padding="same",
                                activation="relu",
                            )
                        )(inputs)
                        x = TimeDistributed(BatchNormalization())(x)
                    else:
                        x = TimeDistributed(
                            Conv2D(
                                filters,
                                kernel_size=(3, 3),
                                padding="same",
                                activation="relu",
                            )
                        )(x)
                        x = TimeDistributed(BatchNormalization())(x)
                else:
                    x = TimeDistributed(
                        Conv2D(
                            filters,
                            kernel_size=(3, 3),
                            padding="same",
                            activation="relu",
                        )
                    )(x)
                    x = TimeDistributed(BatchNormalization())(x)
            x = TimeDistributed(MaxPooling2D(pool_size=(2, 2)))(x)
            filters = int(filters * 2)

        x = TimeDistributed(Flatten())(x)
        x = LSTM(int(filters / 2), return_sequences=False, return_state=False)(x)

        if with_classifier is False:
            return x

        else:
            x = Dense(int(filters / 2), activation="relu")(x)
            x = BatchNormalization()(x)
            x = Dropout(0.5)(x)
            if classes == 2:
                x = Dense(1, activation="sigmoid")(x)
            else:
                x = Dense(classes, activation="softmax")(x)

            model = Model(inputs=inputs, outputs=x)
            # plot_model(model,'model.png',show_shapes=True)

            return model

    def res_block(self, x, filters, strides=2, block=False, basic=True):
        shortcut = x

        if basic is True:
            x = ZeroPadding2D((1, 1))(x)
            x = Conv2D(filters, (3, 3), strides=(strides, strides))(x)

        else:
            x = Conv2D(filters, (1, 1), strides=(strides, strides))(x)

        x = BatchNormalization()(x)
        x = Activation("relu")(x)

        x = ZeroPadding2D((1, 1))(x)
        x = Conv2D(filters, (3, 3), strides=(1, 1))(x)
        x = BatchNormalization()(x)

        if basic is True:
            if block is False:
                shortcut = Conv2D(filters, (1, 1), strides=(strides, strides))(shortcut)
                shortcut = BatchNormalization()(shortcut)

        else:
            x = Activation("relu")(x)

            x = Conv2D(int(filters * 4), (1, 1), strides=(1, 1))(x)
            x = BatchNormalization()(x)

            if block is False:
                shortcut = Conv2D(filters * 4, (1, 1), strides=(strides, strides))(
                    shortcut
                )
                shortcut = BatchNormalization()(shortcut)

        x = Add()([x, shortcut])
        x = Activation("relu")(x)

        return x

    def tres_block(self, x, filters, strides=2, block=False, basic=True):
        shortcut = x

        if basic is True:
            x = TimeDistributed(ZeroPadding2D((1, 1)))(x)
            x = TimeDistributed(Conv2D(filters, (3, 3), strides=(strides, strides)))(x)

        else:
            x = TimeDistributed(Conv2D(filters, (1, 1), strides=(strides, strides)))(x)

        x = TimeDistributed(BatchNormalization())(x)
        x = TimeDistributed(Activation("relu"))(x)

        x = TimeDistributed(ZeroPadding2D((1, 1)))(x)
        x = TimeDistributed(Conv2D(filters, (3, 3), strides=(1, 1)))(x)
        x = TimeDistributed(BatchNormalization())(x)

        if basic is True:
            if block is False:
                shortcut = TimeDistributed(
                    Conv2D(filters, (1, 1), strides=(strides, strides))
                )(shortcut)
                shortcut = TimeDistributed(BatchNormalization())(shortcut)

        else:
            x = TimeDistributed(Activation("relu"))(x)

            x = TimeDistributed(Conv2D(int(filters * 4), (1, 1), strides=(1, 1)))(x)
            x = TimeDistributed(BatchNormalization())(x)

            if block is False:
                shortcut = TimeDistributed(
                    Conv2D(int(filters * 4), (1, 1), strides=(strides, strides))
                )(shortcut)
                shortcut = TimeDistributed(BatchNormalization())(shortcut)

        x = Add()([x, shortcut])
        x = TimeDistributed(Activation("relu"))(x)

        return x

    def simple_resnet(self, inputs, filters, classes=3, level=5, with_classifier=False):
        x = ZeroPadding2D((3, 3))(inputs)
        x = Conv2D(filters, (5, 5), strides=(2, 2))(x)
        x = BatchNormalization()(x)
        x = Activation("relu")(x)
        x = MaxPooling2D((3, 3), strides=(2, 2))(x)

        if level < 6:
            layers = [2, 2, 2, 2]
            basic = True
        elif level < 7:
            layers = [3, 4, 6, 3]
            basic = True
        else:
            layers = [3, 4, 6, 3]
            basic = False

        for i in layers:
            for n in range(i):
                if n == 0:
                    if layers.index(i) == 0:
                        x = self.res_block(
                            x, filters, strides=1, block=False, basic=basic
                        )
                    else:
                        x = self.res_block(
                            x, filters, strides=2, block=False, basic=basic
                        )
                else:
                    x = self.res_block(x, filters, strides=1, block=True, basic=basic)
            filters = int(filters * 2)

        x = AveragePooling2D((2, 2))(x)
        x = Flatten()(x)

        if with_classifier is False:
            return x

        else:
            x = Dropout(0.5)(x)
            if classes == 2:
                x = Dense(1, activation="sigmoid")(x)
            else:
                x = Dense(classes, activation="softmax")(x)

            model = Model(inputs=inputs, outputs=x)

            return model

    def simple_tresnet(
        self, inputs, filters, classes=3, level=5, with_classifier=False
    ):
        x = TimeDistributed(ZeroPadding2D((3, 3)))(inputs)
        x = TimeDistributed(Conv2D(filters, (5, 5), strides=(2, 2)))(x)
        x = TimeDistributed(BatchNormalization())(x)
        x = TimeDistributed(Activation("relu"))(x)
        x = TimeDistributed(MaxPooling2D((3, 3), strides=(2, 2)))(x)

        if level < 6:
            layers = [2, 2, 2, 2]
            basic = True
        elif level < 7:
            layers = [3, 4, 6, 3]
            basic = True
        else:
            layers = [3, 4, 6, 3]
            basic = False

        for i in layers:
            for n in range(i):
                if n == 0:
                    if layers.index(i) == 0:
                        x = self.tres_block(
                            x, filters, strides=1, block=False, basic=basic
                        )
                    else:
                        x = self.tres_block(
                            x, filters, strides=2, block=False, basic=basic
                        )
                else:
                    x = self.tres_block(x, filters, strides=1, block=True, basic=basic)
            filters = int(filters * 2)

        x = TimeDistributed(AveragePooling2D((2, 2)))(x)
        x = TimeDistributed(Flatten())(x)

        if level == 5:
            x = LSTM(1024, return_sequences=False, return_state=False)(x)
        elif level == 6:
            x = LSTM(2048, return_sequences=False, return_state=False)(x)
        else:
            x = LSTM(4096, return_sequences=False, return_state=False)(x)

        if with_classifier is False:
            return x

        else:
            if level == 5:
                x = Dense(1024, activation="relu")(x)
            elif level == 6:
                x = Dense(2048, activation="relu")(x)
            else:
                x = Dense(4096, activation="relu")(x)

            x = Dropout(0.5)(x)
            if classes == 2:
                x = Dense(1, activation="sigmoid")(x)
            else:
                x = Dense(classes, activation="softmax")(x)

            model = Model(inputs=inputs, outputs=x)

            return model

    def combined_network(
        self,
        time_step=15,
        dim_tconv=32,
        dim_conv=64,
        channel=1,
        classes=9,
        level_tconv=1,
        level_conv=2,
    ):
        animation_inputs = Input(shape=(time_step, dim_tconv, dim_tconv, channel))
        pattern_image_inputs = Input(shape=(dim_conv, dim_conv, 3))

        filters_tconv = 8
        filters_conv = 8

        for i in range(round(dim_tconv / 60)):
            filters_tconv = min(int(filters_tconv * 2), 64)

        for i in range(round(dim_conv / 60)):
            filters_conv = min(int(filters_conv * 2), 64)

        if level_tconv < 5:
            animation_feature = self.simple_tvgg(
                animation_inputs,
                filters_tconv,
                level=level_tconv,
                with_classifier=False,
            )
        else:
            animation_feature = self.simple_tresnet(
                animation_inputs,
                filters_tconv,
                level=level_tconv,
                with_classifier=False,
            )

        if level_conv < 5:
            pattern_image_feature = self.simple_vgg(
                pattern_image_inputs,
                filters_conv,
                level=level_conv,
                with_classifier=False,
            )
        else:
            pattern_image_feature = self.simple_resnet(
                pattern_image_inputs,
                filters_conv,
                level=level_conv,
                with_classifier=False,
            )

        merged_features = concatenate([animation_feature, pattern_image_feature])

        nodes = 32
        for i in range(max(level_tconv, level_conv)):
            nodes = int(nodes * 2)
        outputs = Dense(nodes, activation="relu")(merged_features)
        outputs = BatchNormalization()(outputs)
        outputs = Dropout(0.5)(outputs)
        if classes == 2:
            predictions = Dense(1, activation="sigmoid")(outputs)
        else:
            predictions = Dense(classes, activation="softmax")(outputs)

        model = Model(
            inputs=[animation_inputs, pattern_image_inputs], outputs=predictions
        )

        return model

    def train_pattern_recognizer(
        self,
        data_path,
        model_path,
        out_path=None,
        dim=64,
        channel=3,
        time_step=15,
        level=2,
        aug_methods=[],
        augvalid=True,
        include_bodyparts=True,
        std=0,
        background_free=True,
        behavior_mode=0,
        social_distance=0,
    ):
        filters = 8

        for i in range(round(dim / 60)):
            filters = min(int(filters * 2), 64)

        inputs = Input(shape=(dim, dim, channel))

        print(
            "Training the Categorizer w/ only Pattern Recognizer using the behavior examples in: "
            + str(data_path)
        )

        files = [i for i in os.listdir(data_path) if i.endswith(self.extension_image)]

        path_files = []
        labels = []

        for i in files:
            path_file = os.path.join(data_path, i)
            path_files.append(path_file)
            labels.append(os.path.splitext(i)[0].split("_")[-1])

        labels = np.array(labels)
        lb = LabelBinarizer()
        labels = lb.fit_transform(labels)
        self.classnames = lb.classes_

        print("Found behavior names: " + str(self.classnames))

        if include_bodyparts is True:
            inner_code = 0
        else:
            inner_code = 1

        if background_free is True:
            background_code = 0
        else:
            background_code = 1

        if behavior_mode >= 3:
            time_step = std = 0
            inner_code = 1

        parameters = {
            "classnames": list(self.classnames),
            "dim_conv": int(dim),
            "channel": int(channel),
            "time_step": int(time_step),
            "network": 0,
            "level_conv": int(level),
            "inner_code": int(inner_code),
            "std": int(std),
            "background_free": int(background_code),
            "behavior_kind": int(behavior_mode),
            "social_distance": int(social_distance),
        }
        pd_parameters = pd.DataFrame.from_dict(parameters)
        pd_parameters.to_csv(
            os.path.join(model_path, "model_parameters.txt"), index=False
        )

        (train_files, test_files, y1, y2) = train_test_split(
            path_files, labels, test_size=0.2, stratify=labels
        )

        print("Perform augmentation for the behavior examples...")
        print(
            "This might take hours or days, depending on the capacity of your computer."
        )
        print(datetime.datetime.now())

        print("Start to augment training examples...")
        _, trainX, trainY = self.build_data(
            train_files,
            dim_tconv=0,
            dim_conv=dim,
            channel=channel,
            time_step=time_step,
            aug_methods=aug_methods,
            background_free=background_free,
            behavior_mode=behavior_mode,
        )
        trainY = lb.fit_transform(trainY)
        print("Start to augment validation examples...")
        if augvalid is True:
            _, testX, testY = self.build_data(
                test_files,
                dim_tconv=0,
                dim_conv=dim,
                channel=channel,
                time_step=time_step,
                aug_methods=aug_methods,
                background_free=background_free,
                behavior_mode=behavior_mode,
            )
        else:
            _, testX, testY = self.build_data(
                test_files,
                dim_tconv=0,
                dim_conv=dim,
                channel=channel,
                time_step=time_step,
                aug_methods=[],
                background_free=background_free,
                behavior_mode=behavior_mode,
            )
        testY = lb.fit_transform(testY)

        with tf.device("CPU"):
            trainX = tf.convert_to_tensor(trainX)
            trainY = tf.convert_to_tensor(trainY)
            testX_tensor = tf.convert_to_tensor(testX)
            testY_tensor = tf.convert_to_tensor(testY)

        print("Training example shape : " + str(trainX.shape))
        print("Training label shape : " + str(trainY.shape))
        print("Validation example shape : " + str(testX.shape))
        print("Validation label shape : " + str(testY.shape))
        print(datetime.datetime.now())

        if trainX.shape[0] < 5000:
            batch_size = 8
        elif trainX.shape[0] < 50000:
            batch_size = 16
        else:
            batch_size = 32

        if level < 5:
            model = self.simple_vgg(
                inputs,
                filters,
                classes=len(self.classnames),
                level=level,
                with_classifier=True,
            )
        else:
            model = self.simple_resnet(
                inputs,
                filters,
                classes=len(self.classnames),
                level=level,
                with_classifier=True,
            )
        if len(self.classnames) == 2:
            model.compile(
                optimizer=SGD(learning_rate=1e-4, momentum=0.9),
                loss="binary_crossentropy",
                metrics=["accuracy"],
            )
        else:
            model.compile(
                optimizer=SGD(learning_rate=1e-4, momentum=0.9),
                loss="categorical_crossentropy",
                metrics=["accuracy"],
            )

        cp = ModelCheckpoint(
            model_path,
            monitor="val_loss",
            verbose=1,
            save_best_only=True,
            save_weights_only=False,
            mode="min",
            save_freq="epoch",
        )
        es = EarlyStopping(
            monitor="val_loss",
            min_delta=0.001,
            mode="min",
            verbose=1,
            patience=4,
            restore_best_weights=True,
        )
        rl = ReduceLROnPlateau(
            monitor="val_loss",
            min_delta=0.001,
            factor=0.2,
            patience=2,
            verbose=1,
            mode="min",
            min_learning_rate=1e-7,
        )

        H = model.fit(
            trainX,
            trainY,
            batch_size=batch_size,
            validation_data=(testX_tensor, testY_tensor),
            epochs=1000000,
            callbacks=[cp, es, rl],
        )

        model.save(model_path)
        print("Trained Categorizer saved in: " + str(model_path))

        try:
            predictions = model.predict(testX, batch_size=batch_size)

            if len(self.classnames) == 2:
                print(
                    classification_report(
                        testY.argmax(axis=1),
                        predictions.argmax(axis=1),
                        target_names=[self.classnames[0]],
                    )
                )
                report = classification_report(
                    testY.argmax(axis=1),
                    predictions.argmax(axis=1),
                    target_names=[self.classnames[0]],
                    output_dict=True,
                )
            else:
                print(
                    classification_report(
                        testY.argmax(axis=1),
                        predictions.argmax(axis=1),
                        target_names=self.classnames,
                    )
                )
                report = classification_report(
                    testY.argmax(axis=1),
                    predictions.argmax(axis=1),
                    target_names=self.classnames,
                    output_dict=True,
                )

            pd.DataFrame(report).transpose().to_csv(
                os.path.join(model_path, "training_metrics.csv"), float_format="%.2f"
            )
            if out_path is not None:
                pd.DataFrame(report).transpose().to_excel(
                    os.path.join(out_path, "training_metrics.xlsx"), float_format="%.2f"
                )

            plt.style.use("seaborn-bright")
            plt.figure()
            plt.plot(H.history["loss"], label="train_loss")
            plt.plot(H.history["val_loss"], label="val_loss")
            plt.plot(H.history["accuracy"], label="train_accuracy")
            plt.plot(H.history["val_accuracy"], label="val_accuracy")
            plt.title("Loss and Accuracy")
            plt.xlabel("Epoch")
            plt.ylabel("Loss/Accuracy")
            plt.legend(loc="center right")
            plt.savefig(os.path.join(model_path, "training_history.png"))
            if out_path is not None:
                plt.savefig(os.path.join(out_path, "training_history.png"))
                print("Training reports saved in: " + str(out_path))
            plt.close("all")

        except:
            pass

    def train_animation_analyzer(
        self,
        data_path,
        model_path,
        out_path=None,
        dim=64,
        channel=1,
        time_step=15,
        level=2,
        aug_methods=[],
        augvalid=True,
        include_bodyparts=True,
        std=0,
        background_free=True,
        behavior_mode=0,
        social_distance=0,
    ):
        filters = 8

        for i in range(round(dim / 60)):
            filters = min(int(filters * 2), 64)

        inputs = Input(shape=(time_step, dim, dim, channel))

        print(
            "Training the Categorizer w/o Pattern Recognizer using the behavior examples in: "
            + str(data_path)
        )

        files = [i for i in os.listdir(data_path) if i.endswith(self.extension_video)]

        path_files = []
        labels = []

        for i in files:
            path_file = os.path.join(data_path, i)
            path_files.append(path_file)
            labels.append(os.path.splitext(i)[0].split("_")[-1])

        labels = np.array(labels)
        lb = LabelBinarizer()
        labels = lb.fit_transform(labels)
        self.classnames = lb.classes_

        print("Found behavior names: " + str(self.classnames))

        if include_bodyparts is True:
            inner_code = 0
        else:
            inner_code = 1

        if background_free is True:
            background_code = 0
        else:
            background_code = 1

        parameters = {
            "classnames": list(self.classnames),
            "dim_tconv": int(dim),
            "channel": int(channel),
            "time_step": int(time_step),
            "network": 1,
            "level_tconv": int(level),
            "inner_code": int(inner_code),
            "std": int(std),
            "background_free": int(background_code),
            "behavior_kind": int(behavior_mode),
            "social_distance": int(social_distance),
        }
        pd_parameters = pd.DataFrame.from_dict(parameters)
        pd_parameters.to_csv(
            os.path.join(model_path, "model_parameters.txt"), index=False
        )

        (train_files, test_files, y1, y2) = train_test_split(
            path_files, labels, test_size=0.2, stratify=labels
        )

        print("Perform augmentation for the behavior examples...")
        print(
            "This might take hours or days, depending on the capacity of your computer."
        )
        print(datetime.datetime.now())

        print("Start to augment training examples...")
        trainX, _, trainY = self.build_data(
            train_files,
            dim_tconv=dim,
            dim_conv=dim,
            channel=channel,
            time_step=time_step,
            aug_methods=aug_methods,
            background_free=background_free,
            behavior_mode=behavior_mode,
        )
        trainY = lb.fit_transform(trainY)
        print("Start to augment validation examples...")
        if augvalid is True:
            testX, _, testY = self.build_data(
                test_files,
                dim_tconv=dim,
                dim_conv=dim,
                channel=channel,
                time_step=time_step,
                aug_methods=aug_methods,
                background_free=background_free,
                behavior_mode=behavior_mode,
            )
        else:
            testX, _, testY = self.build_data(
                test_files,
                dim_tconv=dim,
                dim_conv=dim,
                channel=channel,
                time_step=time_step,
                aug_methods=[],
                background_free=background_free,
                behavior_mode=behavior_mode,
            )
        testY = lb.fit_transform(testY)

        with tf.device("CPU"):
            trainX = tf.convert_to_tensor(trainX)
            trainY = tf.convert_to_tensor(trainY)
            testX_tensor = tf.convert_to_tensor(testX)
            testY_tensor = tf.convert_to_tensor(testY)

        print("Training example shape : " + str(trainX.shape))
        print("Training label shape : " + str(trainY.shape))
        print("Validation example shape : " + str(testX.shape))
        print("Validation label shape : " + str(testY.shape))
        print(datetime.datetime.now())

        if trainX.shape[0] < 5000:
            batch_size = 8
        elif trainX.shape[0] < 50000:
            batch_size = 16
        else:
            batch_size = 32

        if level < 5:
            model = self.simple_tvgg(
                inputs,
                filters,
                classes=len(self.classnames),
                level=level,
                with_classifier=True,
            )
        else:
            model = self.simple_tresnet(
                inputs,
                filters,
                classes=len(self.classnames),
                level=level,
                with_classifier=True,
            )

        if len(self.classnames) == 2:
            model.compile(
                optimizer=SGD(learning_rate=1e-4, momentum=0.9),
                loss="binary_crossentropy",
                metrics=["accuracy"],
            )
        else:
            model.compile(
                optimizer=SGD(learning_rate=1e-4, momentum=0.9),
                loss="categorical_crossentropy",
                metrics=["accuracy"],
            )

        cp = ModelCheckpoint(
            model_path,
            monitor="val_loss",
            verbose=1,
            save_best_only=True,
            save_weights_only=False,
            mode="min",
            save_freq="epoch",
        )
        es = EarlyStopping(
            monitor="val_loss",
            min_delta=0.001,
            mode="min",
            verbose=1,
            patience=4,
            restore_best_weights=True,
        )
        rl = ReduceLROnPlateau(
            monitor="val_loss",
            min_delta=0.001,
            factor=0.2,
            patience=2,
            verbose=1,
            mode="min",
            min_learning_rate=1e-7,
        )

        H = model.fit(
            trainX,
            trainY,
            batch_size=batch_size,
            validation_data=(testX_tensor, testY_tensor),
            epochs=1000000,
            callbacks=[cp, es, rl],
        )

        model.save(model_path)
        print("Trained Categorizer saved in: " + str(model_path))

        try:
            predictions = model.predict(testX, batch_size=batch_size)

            if len(self.classnames) == 2:
                print(
                    classification_report(
                        testY.argmax(axis=1),
                        predictions.argmax(axis=1),
                        target_names=[self.classnames[0]],
                    )
                )
                report = classification_report(
                    testY.argmax(axis=1),
                    predictions.argmax(axis=1),
                    target_names=[self.classnames[0]],
                    output_dict=True,
                )
            else:
                print(
                    classification_report(
                        testY.argmax(axis=1),
                        predictions.argmax(axis=1),
                        target_names=self.classnames,
                    )
                )
                report = classification_report(
                    testY.argmax(axis=1),
                    predictions.argmax(axis=1),
                    target_names=self.classnames,
                    output_dict=True,
                )

            pd.DataFrame(report).transpose().to_csv(
                os.path.join(model_path, "training_metrics.csv"), float_format="%.2f"
            )
            if out_path is not None:
                pd.DataFrame(report).transpose().to_excel(
                    os.path.join(out_path, "training_metrics.xlsx"), float_format="%.2f"
                )

            plt.style.use("seaborn-bright")
            plt.figure()
            plt.plot(H.history["loss"], label="train_loss")
            plt.plot(H.history["val_loss"], label="val_loss")
            plt.plot(H.history["accuracy"], label="train_accuracy")
            plt.plot(H.history["val_accuracy"], label="val_accuracy")
            plt.title("Loss and Accuracy")
            plt.xlabel("Epoch")
            plt.ylabel("Loss/Accuracy")
            plt.legend(loc="center right")
            plt.savefig(os.path.join(model_path, "training_history.png"))
            if out_path is not None:
                plt.savefig(os.path.join(out_path, "training_history.png"))
                print("Training reports saved in: " + str(out_path))
            plt.close("all")

        except:
            pass

    def train_combnet(
        self,
        data_path,
        model_path,
        out_path=None,
        dim_tconv=32,
        dim_conv=64,
        channel=1,
        time_step=15,
        level_tconv=1,
        level_conv=2,
        aug_methods=[],
        augvalid=True,
        include_bodyparts=True,
        std=0,
        background_free=True,
        behavior_mode=0,
        social_distance=0,
    ):
        print(
            "Training Categorizer with both Animation Analyzer and Pattern Recognizer using the behavior examples in: "
            + str(data_path)
        )

        files = [i for i in os.listdir(data_path) if i.endswith(self.extension_video)]

        path_files = []
        labels = []

        for i in files:
            path_file = os.path.join(data_path, i)
            path_files.append(path_file)
            labels.append(os.path.splitext(i)[0].split("_")[-1])

        labels = np.array(labels)
        lb = LabelBinarizer()
        labels = lb.fit_transform(labels)
        self.classnames = lb.classes_

        print("Found behavior names: " + str(self.classnames))

        if include_bodyparts is True:
            inner_code = 0
        else:
            inner_code = 1

        if background_free is True:
            background_code = 0
        else:
            background_code = 1

        parameters = {
            "classnames": list(self.classnames),
            "dim_tconv": int(dim_tconv),
            "dim_conv": int(dim_conv),
            "channel": int(channel),
            "time_step": int(time_step),
            "network": 2,
            "level_tconv": int(level_tconv),
            "level_conv": int(level_conv),
            "inner_code": int(inner_code),
            "std": int(std),
            "background_free": int(background_code),
            "behavior_kind": int(behavior_mode),
            "social_distance": int(social_distance),
        }
        pd_parameters = pd.DataFrame.from_dict(parameters)
        pd_parameters.to_csv(
            os.path.join(model_path, "model_parameters.txt"), index=False
        )

        (train_files, test_files, y1, y2) = train_test_split(
            path_files, labels, test_size=0.2, stratify=labels
        )

        print("Perform augmentation for the behavior examples...")
        print(
            "This might take hours or days, depending on the capacity of your computer."
        )
        print(datetime.datetime.now())

        print("Start to augment training examples...")
        train_animations, train_pattern_images, trainY = self.build_data(
            train_files,
            dim_tconv=dim_tconv,
            dim_conv=dim_conv,
            channel=channel,
            time_step=time_step,
            aug_methods=aug_methods,
            background_free=background_free,
            behavior_mode=behavior_mode,
        )
        trainY = lb.fit_transform(trainY)
        print("Start to augment validation examples...")
        if augvalid is True:
            test_animations, test_pattern_images, testY = self.build_data(
                test_files,
                dim_tconv=dim_tconv,
                dim_conv=dim_conv,
                channel=channel,
                time_step=time_step,
                aug_methods=aug_methods,
                background_free=background_free,
                behavior_mode=behavior_mode,
            )
        else:
            test_animations, test_pattern_images, testY = self.build_data(
                test_files,
                dim_tconv=dim_tconv,
                dim_conv=dim_conv,
                channel=channel,
                time_step=time_step,
                aug_methods=[],
                background_free=background_free,
                behavior_mode=behavior_mode,
            )
        testY = lb.fit_transform(testY)

        with tf.device("CPU"):
            train_animations = tf.convert_to_tensor(train_animations)
            train_pattern_images = tf.convert_to_tensor(train_pattern_images)
            trainY = tf.convert_to_tensor(trainY)
            test_animations_tensor = tf.convert_to_tensor(test_animations)
            test_pattern_images_tensor = tf.convert_to_tensor(test_pattern_images)
            testY_tensor = tf.convert_to_tensor(testY)

        print(
            "Training example shape : "
            + str(train_animations.shape)
            + ", "
            + str(train_pattern_images.shape)
        )
        print("Training label shape : " + str(trainY.shape))
        print(
            "Validation example shape : "
            + str(test_animations.shape)
            + ", "
            + str(test_pattern_images.shape)
        )
        print("Validation label shape : " + str(testY.shape))
        print(datetime.datetime.now())

        if train_animations.shape[0] < 5000:
            batch_size = 8
        elif train_animations.shape[0] < 50000:
            batch_size = 16
        else:
            batch_size = 32

        model = self.combined_network(
            time_step=time_step,
            dim_tconv=dim_tconv,
            dim_conv=dim_conv,
            channel=channel,
            classes=len(self.classnames),
            level_tconv=level_tconv,
            level_conv=level_conv,
        )
        if len(self.classnames) == 2:
            model.compile(
                optimizer=SGD(learning_rate=1e-4, momentum=0.9),
                loss="binary_crossentropy",
                metrics=["accuracy"],
            )
        else:
            model.compile(
                optimizer=SGD(learning_rate=1e-4, momentum=0.9),
                loss="categorical_crossentropy",
                metrics=["accuracy"],
            )

        cp = ModelCheckpoint(
            model_path,
            monitor="val_loss",
            verbose=1,
            save_best_only=True,
            save_weights_only=False,
            mode="min",
            save_freq="epoch",
        )
        es = EarlyStopping(
            monitor="val_loss",
            min_delta=0.001,
            mode="min",
            verbose=1,
            patience=4,
            restore_best_weights=True,
        )
        rl = ReduceLROnPlateau(
            monitor="val_loss",
            min_delta=0.001,
            factor=0.2,
            patience=2,
            verbose=1,
            mode="min",
            min_learning_rate=1e-7,
        )

        H = model.fit(
            [train_animations, train_pattern_images],
            trainY,
            batch_size=batch_size,
            validation_data=(
                [test_animations_tensor, test_pattern_images_tensor],
                testY_tensor,
            ),
            epochs=1000000,
            callbacks=[cp, es, rl],
        )

        model.save(model_path)
        print("Trained Categorizer saved in: " + str(model_path))

        try:
            predictions = model.predict(
                [test_animations, test_pattern_images], batch_size=batch_size
            )

            if len(self.classnames) == 2:
                print(
                    classification_report(
                        testY.argmax(axis=1),
                        predictions.argmax(axis=1),
                        target_names=[self.classnames[0]],
                    )
                )
                report = classification_report(
                    testY.argmax(axis=1),
                    predictions.argmax(axis=1),
                    target_names=[self.classnames[0]],
                    output_dict=True,
                )
            else:
                print(
                    classification_report(
                        testY.argmax(axis=1),
                        predictions.argmax(axis=1),
                        target_names=self.classnames,
                    )
                )
                report = classification_report(
                    testY.argmax(axis=1),
                    predictions.argmax(axis=1),
                    target_names=self.classnames,
                    output_dict=True,
                )

            pd.DataFrame(report).transpose().to_csv(
                os.path.join(model_path, "training_metrics.csv"), float_format="%.2f"
            )
            if out_path is not None:
                pd.DataFrame(report).transpose().to_excel(
                    os.path.join(out_path, "training_metrics.xlsx"), float_format="%.2f"
                )

            plt.style.use("seaborn-bright")
            plt.figure()
            plt.plot(H.history["loss"], label="train_loss")
            plt.plot(H.history["val_loss"], label="val_loss")
            plt.plot(H.history["accuracy"], label="train_accuracy")
            plt.plot(H.history["val_accuracy"], label="val_accuracy")
            plt.title("Loss and Accuracy")
            plt.xlabel("Epoch")
            plt.ylabel("Loss/Accuracy")
            plt.legend(loc="center right")
            plt.savefig(os.path.join(model_path, "training_history.png"))
            if out_path is not None:
                plt.savefig(os.path.join(out_path, "training_history.png"))
                print("Training reports saved in: " + str(out_path))
            plt.close("all")

        except:
            pass

    def test_categorizer(self, groundtruth_path, model_path, result_path=None):
        print("Testing the selected Categorizer...")

        animations = deque()
        pattern_images = deque()
        labels = deque()

        parameters = pd.read_csv(os.path.join(model_path, "model_parameters.txt"))

        if "dim_conv" in list(parameters.keys()):
            dim_conv = int(parameters["dim_conv"][0])
        if "dim_tconv" in list(parameters.keys()):
            dim_tconv = int(parameters["dim_tconv"][0])
        if "level_conv" in list(parameters.keys()):
            level_conv = int(parameters["level_conv"][0])
        if "dim_tconv" in list(parameters.keys()):
            level_tconv = int(parameters["level_tconv"][0])
        if "channel" in list(parameters.keys()):
            channel = int(parameters["channel"][0])
        network = int(parameters["network"][0])
        if network == 0:
            print(
                "The type of the Categorizer: Pattern Recognizer (Lv "
                + str(level_conv)
                + "; Shape "
                + str(dim_conv)
                + " X "
                + str(dim_conv)
                + " X 3)."
            )
        if network == 2:
            print(
                "The type of the Categorizer: Animation Analyzer (Lv "
                + str(level_tconv)
                + "; Shape "
                + str(dim_tconv)
                + " X "
                + str(dim_tconv)
                + " X "
                + str(channel)
                + ") + Pattern Recognizer (Lv "
                + str(level_conv)
                + "; Shape "
                + str(dim_conv)
                + " X "
                + str(dim_conv)
                + " X 3)."
            )
        if "behavior_kind" in list(parameters.keys()):
            behavior_mode = int(parameters["behavior_kind"][0])
        else:
            behavior_mode = 0
        if behavior_mode == 0:
            print("The behavior mode of the Categorizer: Non-interactive.")
        elif behavior_mode == 1:
            print("The behavior mode of the Categorizer: Interactive basic.")
        elif behavior_mode == 2:
            print(
                "The behavior mode of the Categorizer: Interactive advanced (Social distance "
                + str(parameters["social_distance"][0])
                + ")."
            )
        else:
            print(
                "The behavior mode of the Categorizer: Static images (non-interactive)."
            )
        length = int(parameters["time_step"][0])
        print(
            "The length of a behavior example in the Categorizer: "
            + str(length)
            + " frames."
        )
        if int(parameters["inner_code"][0]) == 0:
            print(
                "The Categorizer includes body parts in analysis with STD = "
                + str(parameters["std"][0])
                + "."
            )
        else:
            print("The Categorizer does not include body parts in analysis.")
        if int(parameters["background_free"][0]) == 0:
            print("The Categorizer does not include background in analysis.")
        else:
            print("The Categorizer includes background in analysis.")
        classnames = list(parameters["classnames"])
        print("Behavior names in the Categorizer: " + str(classnames))
        behaviornames = [
            i
            for i in os.listdir(groundtruth_path)
            if os.path.isdir(os.path.join(groundtruth_path, i))
        ]
        incorrect_behaviors = list(set(behaviornames) - set(classnames))
        incorrect_classes = list(set(classnames) - set(behaviornames))
        if len(incorrect_behaviors) > 0:
            print(
                "Mismatched behavior names in testing examples: "
                + str(incorrect_behaviors)
            )
        if len(incorrect_classes) > 0:
            print("Unused behavior names in the Categorizer: " + str(incorrect_classes))

        if len(incorrect_behaviors) == 0 and len(incorrect_classes) == 0:
            for behavior in behaviornames:
                if network != 0:
                    filenames = [
                        i
                        for i in os.listdir(os.path.join(groundtruth_path, behavior))
                        if i.endswith(".avi")
                    ]
                else:
                    filenames = [
                        i
                        for i in os.listdir(os.path.join(groundtruth_path, behavior))
                        if i.endswith(".jpg")
                    ]

                for i in filenames:
                    if network != 0:
                        path_to_animation = os.path.join(groundtruth_path, behavior, i)

                        capture = cv2.VideoCapture(path_to_animation)
                        animation = deque()
                        frames = deque(maxlen=length)

                        while True:
                            retval, frame = capture.read()
                            if frame is None:
                                break
                            frames.append(frame)

                        capture.release()

                        for frame in frames:
                            frame = np.uint8(
                                exposure.rescale_intensity(frame, out_range=(0, 255))
                            )
                            if channel == 1:
                                frame = cv2.cvtColor(
                                    np.uint8(frame), cv2.COLOR_BGR2GRAY
                                )
                            frame = cv2.resize(
                                frame,
                                (dim_tconv, dim_tconv),
                                interpolation=cv2.INTER_AREA,
                            )
                            frame = img_to_array(frame)
                            animation.append(frame)

                        animations.append(np.array(animation))

                    if network != 1:
                        path_to_pattern_image = (
                            os.path.splitext(
                                os.path.join(groundtruth_path, behavior, i)
                            )[0]
                            + ".jpg"
                        )
                        pattern_image = cv2.imread(path_to_pattern_image)
                        pattern_image = cv2.resize(
                            pattern_image,
                            (dim_conv, dim_conv),
                            interpolation=cv2.INTER_AREA,
                        )
                        pattern_images.append(img_to_array(pattern_image))

                    labels.append(classnames.index(behavior))

            if network != 0:
                animations = np.array(animations, dtype="float32") / 255.0
            pattern_images = np.array(pattern_images, dtype="float32") / 255.0

            labels = np.array(labels)

            model = load_model(model_path)

            if network == 0:
                predictions = model.predict(pattern_images, batch_size=32)
            elif network == 1:
                predictions = model.predict(animations, batch_size=32)
            else:
                predictions = model.predict([animations, pattern_images], batch_size=32)

            print(
                classification_report(
                    labels, predictions.argmax(axis=1), target_names=classnames
                )
            )
            report = classification_report(
                labels,
                predictions.argmax(axis=1),
                target_names=classnames,
                output_dict=True,
            )

            if result_path is not None:
                pd.DataFrame(report).transpose().to_excel(
                    os.path.join(result_path, "testing_reports.xlsx"),
                    float_format="%.2f",
                )

            print("Testing completed!")



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""

from typing import Callable

import wx
import wx.lib.agw.hyperlink

from .behavior_examples import GenerateBehaviorExamples, SortBehaviorExamples
from .categorizers import (
    TestCategorizers,
    TrainCategorizers,
)
from .detectors import (
    GenerateImageExamples,
    TestDetectors,
    TrainDetectors,
)


class TrainingModule(wx.Frame):
    """Contains functions related to training and testing Detectors and Categorizers."""

    def __init__(self):
        super().__init__(parent=None, title="Training Module", size=(500, 560))
        self.panel = wx.Panel(self)
        self.boxsizer = wx.BoxSizer(wx.VERTICAL)
        self.boxsizer.Add(0, 60, 0)

        self.add_button(
            "Generate Image Examples",
            self.generate_images,
            "Extract frames from videos for annotating animals / objects in them so that they can be used to train a Detector to detect animals / objects of your interest. See Extended Guide for how to select images to annotate.",
        )

        annotate_link = wx.lib.agw.hyperlink.HyperLinkCtrl(
            self.panel,
            0,
            "\nAnnotate images with Roboflow\n",
            URL="https://roboflow.com",
        )
        self.boxsizer.Add(annotate_link, 0, wx.ALIGN_CENTER, 10)
        self.boxsizer.Add(0, 5, 0)

        self.add_button(
            "Train Detectors",
            self.train_detectors,
            "There are two detection methods in LabGym, the Detector-based method is more versatile (useful in any recording conditions and complex interactive behaviors) but slower than the other background subtraction-based method (requires static background and stable illumination in videos).",
        )

        self.add_button(
            "Test Detectors",
            self.test_detectors,
            "Test trained Detectors on the annotated ground-truth image dataset (similar to the image dataset used for training a Detector).",
        )

        self.boxsizer.Add(0, 45, 0)

        self.add_button(
            "Generate Behavior Examples",
            self.generate_behavior_examples,
            "Generate behavior examples for sorting them so that they can be used to teach a Categorizer to recognize behaviors defined by you.",
        )

        self.add_button(
            "Sort Behavior Examples",
            self.sort_behavior_examples,
            "Set shortcut keys for behavior categories to help sorting the behavior examples in an easier way. See Extended Guide for how to select and sort the behavior examples.",
        )

        self.add_button(
            "Train Categorizers",
            self.train_categorizers,
            "Customize a Categorizer and use the sorted behavior examples to train it so that it can recognize the behaviors of your interest during analysis.",
        )

        self.add_button(
            "Test Categorizers",
            self.test_categorizers,
            "Test trained Categorizers on the sorted ground-truth behavior examples (similar to the behavior examples used for training a Categorizer).",
        )

        self.boxsizer.Add(0, 45, 0)
        self.panel.SetSizer(self.boxsizer)
        self.Centre()
        self.Show(True)

    def add_button(
        self, button_label: str, button_handler: Callable, button_tool_tip: str
    ):
        """
        Adds a button to the main sizer.

        Args:
            button_label: The button label.
            button_handler: The function to handle the button press.
            button_tool_tip: The text displayed when the user hovers over the button.
        """
        button = wx.Button(self.panel, label=button_label, size=(300, 40))
        button.Bind(wx.EVT_BUTTON, button_handler)
        wx.Button.SetToolTip(button, button_tool_tip)
        self.boxsizer.Add(button, 0, wx.ALIGN_CENTER, 10)
        self.boxsizer.Add(0, 5, 0)

    def generate_images(self, event):
        GenerateImageExamples()

    def train_detectors(self, event):
        TrainDetectors()

    def test_detectors(self, event):
        TestDetectors()

    def generate_behavior_examples(self, event):
        GenerateBehaviorExamples()

    def sort_behavior_examples(self, event):
        SortBehaviorExamples()

    def train_categorizers(self, event):
        TrainCategorizers()

    def test_categorizers(self, event):
        TestCategorizers()
